/*
 *
 * @(#) Adapted (by PLI) from: EventMonitor.java 1.10@(#)
 *
 * Copyright (c) 1997 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 * 
 */

/**
 * <p>
 * sieve.beanbox.encapsulatedEvents.EventMonitor.java
 * </p>
 *
 * @version 1.0
 * @author Laurence P. G. Cable.
 */

package sieve.beanbox.encapsulatedEvents;

import java.awt.*;
import java.beans.*;
import java.util.EventObject;
import java.util.Hashtable;
import java.util.Vector;

/*
 * Adapted from the EventManager (private) class used by EventMonitor
 */

public class NewEventManager implements EncapsulatedEventManager {
    private int		 	eventCnt = 1;

    private transient Hashtable	sources = new Hashtable();


    /**
     * <p>
     * Called to determine the events that a particular source emits. Will throw
     * IllegalArgumentException if s is not managed by this
     * EncapsulatedEventManager.
     * </p>
     *
     * @param s the event source
     *
     * @return list of java.util.EventListener sub-interfaces that s sources or null.
     *
     * @throws IllegalArgumentException
     * @throws NullPointerException
     */

    public Class[] getSourceEventListenerInterfaces(Object s) {
	EventSetDescriptor[] esd = null;

	try {
	    esd = Introspector.getBeanInfo(s.getClass()).getEventSetDescriptors();
	} catch (IntrospectionException ie) {
	    System.err.println("Failed to introspect");
	}

	Class[]		     ifs = (esd != null && esd.length >  0
					? new Class[esd.length]
					: null
			     );
	
	for (int i = 0; i < ifs.length; i++) {
	    ifs[i] = esd[i].getListenerType();
	}

	return ifs;
    }

    /**
     * <p>
     * Adds the listener eel to the source s to receive ALL events s emits
     * as encpasulated events. Will throw IllegalArgumentException if s is
     * not managed by this EncapsulatedEventManager.
     * </p>
     *
     * @param s		the event source
     * @param eel	the listener
     *
     * @throws IllegalArgumentException
     * @throws NullPointerException
     */

    public void addEncapsulatedEventListener(Object s, EncapsulatedEventListener eel) {
	addEncapsulatedEventListener(s, eel, getSourceEventListenerInterfaces(s));
    }

    /**
     * <p>
     * Removes the listener eel from the source s, thus unregistering eel from
     * receiving encapsulated events for all the events that s emits.
     * Will throw IllegalArgumentException if s is not managed by this
     * EncapsulatedEventManager.
     * </p>
     *
     * @param s		the event source
     * @param eel	the listener
     *
     * @throws IllegalArgumentException
     * @throws NullPointerException
     */

    public void removeEncapsulatedEventListener(Object s, EncapsulatedEventListener eel) {
	removeEncapsulatedEventListener(s, eel, getSourceEventListenerInterfaces(s));
    }

    /**
     * <p>
     * Adds the listener eel to the source s to receive the events enumerated
     * by lc that s emits. Will throw IllegalArgumentException if s is not
     * managed by this EncapsulatedEventManager or if lc contains a reference
     * to a Class that s does not emit events on.
     * </p>
     *
     * @param s		the event source
     * @param eel	the listener
     * @param lc	the list of events to register the eel to receive.
     *
     * @throws IllegalArgumentException
     * @throws NullPointerException
     */

    public synchronized void addEncapsulatedEventListener(
				Object       		  s,
				EncapsulatedEventListener eel,
				Class[]		          lc
    ) {
	if (s   == null) throw new NullPointerException("Source");
	if (eel == null) throw new NullPointerException("Listener");
	if (lc  == null) throw new NullPointerException("Interfaces");

	Hashtable interfaces = (Hashtable)sources.get(s);

	if (interfaces == null) {
	    interfaces = new Hashtable();
	    sources.put(s, interfaces);
	}

	for (int i = 0; i < lc.length; i++) {
	    EncapsulatedEventAdaptor eea = (EncapsulatedEventAdaptor)interfaces.get(lc[i]);

	    if (eea == null) { // its new
		try {
		    eea = EncapsulatedEventAdaptor.getEncapsulatedEventAdaptor(lc[i], s);
		} catch (ClassNotFoundException  cnfe)  {
		    System.err.println("Cant create adaptor class");
		    cnfe.printStackTrace();
		} catch (InstantiationException  inste) {
		    System.err.println("Cant instantiate adaptor");
		} catch (IllegalAccessException iae)   {
		    System.err.println("illegal access");
		} catch (IntrospectionException  intre) {
		    System.err.println("introspection error");
		} catch (java.lang.NoClassDefFoundError ncdfe) {
		    System.err.println("Got NoClassDefFoundError for " +
			lc[i] + " and " + s);
		    ncdfe.printStackTrace();
		}

		if (eea != null) {
		    synchronized(eea) {
			eea.addEncapsulatedEventListener(eel);
		    }
    		    interfaces.put(lc[i], eea);

// System.out.println("Listening to: " + lc[i].getName() + " on: " + s + "\n");
		} else {
    		    // PLI interfaces.put(lc[i], null);
		}
	    }
	}
    }

    /**
     * <p>
     * Removes the listener eel from the source s, thus unregistering for 
     * encapsulated events from s enumerated by lc. Will throw
     * IllegalArgumentException if s is not managed by this
     * EncapsulatedEventManager or if lc contains a reference
     * to a Class that s does not emit events on. 
     * </p>
     *
     * @param s		the event source
     * @param eel	the listener
     * @param lc	the list of events to unregister the eel from receiving.
     *
     * @throws IllegalArgumentException
     * @throws NullPointerException
     */

    public synchronized void removeEncapsulatedEventListener(
				Object	     	  	  s,
				EncapsulatedEventListener eel,
    			        Class[]			  lc
    ) {

	if (s   == null) throw new NullPointerException("Source");
	if (eel == null) throw new NullPointerException("Listener");
	if (lc  == null) throw new NullPointerException("Interfaces");

	Hashtable interfaces = (Hashtable)sources.get(s);

	
	if (interfaces == null) throw new IllegalArgumentException("Source");

	for (int i = 0; i < lc.length; i++) {
	    EncapsulatedEventAdaptor eea = (EncapsulatedEventAdaptor)interfaces.get(lc[i]);

	    if (eea != null) {
		eea.removeEncapsulatedEventListener(eel);

		synchronized(eea) {
		    if (eea.getEncapsulatedListenerCount() == 0) try {
			eea.setSource(null);
		    } catch (Exception e) {
			System.err.println("failed to unregister source");
		    }

System.out.println("No longer listening to: " + lc[i].getName() + " on: " + s + "\n");
		    interfaces.put(lc[i], null); 
		} 
	    }
	}
    }

}
