package sieve.nuggets.kit;

import java.util.Date;

/**
 * Linearly distributes a set of Dates on an axis. 
 * This needs a way to specify units and a way to optionally skip weekends.
 */
public class DateDataDistributor implements DataDistributor {
    private Date first = null;
    private Date last = null;
    private LinearDataDistributor ldd = null;
    private int unit = 1000 * 60 * 60 * 24;
    private long length = 0;
    private boolean includesWeekends = false;

    public DateDataDistributor() { 
	ldd = new LinearDataDistributor();
    }

    public boolean getIncludesWeekends() {
	return includesWeekends;
    }
    public void setIncludesWeekends(boolean includesWeekends) {
	this.includesWeekends = includesWeekends;
	resetMinMax();
    }

    private long getDistance(Date d) {
	long length = 0;
	if ((first != null) && (d != null)) 
	    length = (d.getTime() - first.getTime())/unit;
	return length;
    }

    private void resetMinMax() {
	if ((first != null) && (last != null)) {

	    length = getDistance(last);
	    long weeks = length/7;
	    long extra = length%7;

	    ldd.setMin(first.getTime()/unit);
	    ldd.setMax(first.getTime()/unit + weeks * 5 + extra);
	}
    }

    /**
     * Sets the first date.
     */
    public void setFirstDate(Date first) {
	this.first = first;
	resetMinMax();
    }

    public Date getFirstDate() {
	return first;
    }

    public long getUnit() {
	return unit;
    }

    /**
     * Sets the last date.
     */
    public void setLastDate(Date last) {
	this.last = last;
	resetMinMax();
    }

    public Date getLastDate() {
	return last;
    }

    /**
     * Returns the size (in pixels) available for representation of the given
     * datapoint.
     */
    public int getMaxSize(Object dataPoint) {
	if (dataPoint instanceof Date) {
	    long offset = 0;
	    if (! includesWeekends) {
		Date d = (Date) dataPoint;
	        long weeks = (d.getTime()/unit - (d.getDay() - first.getDay()) -
		    first.getTime()/unit)/7;
		offset = weeks * 2;
	    }
	    return ldd.getMaxSize(
		    new Long(((Date)dataPoint).getTime()/unit - offset));
	}
	else
	    return -Integer.MAX_VALUE;
    }

    /**
     * Return the pixel in the center of the space available for representation
     * of the given datapoint.
     */
    public int getCentralPixel(Object dataPoint) {
	if (dataPoint instanceof Date) {
	    Date d = (Date) dataPoint;
	    long totalDays = getDistance((Date)dataPoint);
	    if (! includesWeekends) {
		int day = d.getDay();
		if (day == 0 || day == 6)
	 	    return -Integer.MAX_VALUE;
	        long weeks = (d.getTime()/unit - (day - first.getDay()) -
		    first.getTime()/unit)/7;
		totalDays = totalDays - 2 * weeks;
	    }
	    return ldd.getCentralPixel(
		new Long(first.getTime()/unit + totalDays));
	}
	return -Integer.MAX_VALUE;
    }

    public Object getDataPoint(int pixel) {
	Object n = ldd.getDataPoint(pixel);
	if ((n != null) && (n instanceof Number)) {
	    long days = ((Number)n).longValue();
	    if (! includesWeekends) {
	        days += 2 * 
		    (((days - first.getTime()/unit) + first.getDay() - 1)/5);
	    }
	    // Need to look up locale stuff in 1.1
	    return new Date(days * unit + (unit * 7)/24);
	}
	return null;
    }
 
    /**
     * Sets the number of pixels over which data points may be distributed.
     */
    public void setSize(int pixels) {
	ldd.setSize(pixels);
    }

}

