package sieve.phys.circuit;
import java.awt.*;
import java.awt.event.*;
import sieve.workbench.*;
import java.io.Serializable;
import sieve.workbench.event.*;

import java.util.*;

/**
 * This class registers itself with a Workbench object and listens for
 * interesting things to happen. When CircuitComponents are added,
 * removed, linked, or unlinked, the appropriate calls are (will be)
 * made on the Overlord class.
 */
public class CircuitWorkbenchListener extends WorkbenchListenerAdapter {
    /**
     * There may be multiple workbenches, so we keep track of an instance
     * of this class for each in this static table.
     */
    private static Hashtable instances = new Hashtable();

    /**
     * The workbench this object is currently listening to.
     */ 
    private Workbench wb = null;

    /**
     * The "Overlord" object which manages the internal representation of
     * the circuit.
     */
    private Overlord circuit = null;
    private Nodes N1, N2 = null;
    /**
     * Constructs a new CircuitWorkbenchListener. This constructor is
     * protected because we want to control the construction of these --
     * there should only be one per workbench instance.
     */
    protected CircuitWorkbenchListener(Workbench wb) {
	this.wb = wb;
	wb.addWorkbenchListener(this);
	circuit = new Overlord();
    }

    /**
     * Returns the CircuitWorkbenchListener associated with the given 
     * workbench. This is called by CircuitComponentTool. Currently
     * other objects should probably not try to do anything with the
     * CircuitWorkbenchListener instance that this method returns -- it
     * primarily serves as a way to make sure that CircuitWorkbenchListeners
     * are properly constructed.
     */
    public static CircuitWorkbenchListener 
	getCircuitWorkbenchListener(Workbench wb) {
	// See if one of these already exists for these workbenches
	CircuitWorkbenchListener l = 
	    (CircuitWorkbenchListener) instances.get(wb);
	// Otherwise create one and add it to the table
	if (l == null) {
	    l = new CircuitWorkbenchListener(wb);
	    instances.put(wb, l);
	}
	return l;
    }

    /**
     * Called when a link is added.
     */
    public void linkAdded(LinkEvent e) {
	Link l = e.getLink();
	WorkbenchComponent src = l.getSourceComponent();
	WorkbenchComponent dest = l.getDestComponent();
	// Make sure both of these are CircuitComponents
	if ((src instanceof CircuitComponent) && 
	    (dest instanceof CircuitComponent) &&
	    (l instanceof CircuitComponentLink)) {
	    CircuitComponentLink ccl = (CircuitComponentLink) l;
	    circuit.addNodes(ccl.getSourceTerminal().getID(),
		ccl.getDestTerminal().getID());
	    circuit.findLoops();
        }
    }

    /**
     * Called when a link is removed.
     */
    public void linkRemoved(LinkEvent e) {
	Link l = e.getLink();
	WorkbenchComponent src = l.getSourceComponent();
	WorkbenchComponent dest = l.getDestComponent();
	// Make sure both of these are CircuitComponents
	if ((src instanceof CircuitComponent) && 
	    (dest instanceof CircuitComponent) &&
	    (l instanceof CircuitComponentLink)) {
	    CircuitComponentLink ccl = (CircuitComponentLink) l;
	    circuit.removeNodes(ccl.getSourceTerminal().getID(),
		ccl.getDestTerminal().getID());
	    circuit.findLoops();
	}
    }

    /**
     * Called when a component is added to the workbench. 
     */
    public void componentAdded(WorkbenchComponentEvent e) {
	WorkbenchComponent wbc = e.getWorkbenchComponent();
	// Make sure this is a CircuitComponent
        if (wbc instanceof Nodes){
        }
        else
	if (wbc instanceof CircuitComponent) {
	    Terminal[] t = ((CircuitComponent)wbc).getTerminals();
	    circuit.addComponent((CircuitComponent)wbc, 
		t[0].getID(), t[1].getID());
	    circuit.findLoops();
	}
    }

    /**
     * Called when a component is removed from the workbench.
     */
    public void componentRemoved(WorkbenchComponentEvent e) {
	WorkbenchComponent wbc = e.getWorkbenchComponent();
	// Make sure this is a CircuitComponent
	if (wbc instanceof CircuitComponent) {
            wb.setCurrentTool(null);
	    Terminal[] t = ((CircuitComponent)wbc).getTerminals();
            circuit.removeNodes(t[0].getID(), t[1].getID());
            if( t[0].numLinks() >1 && !(wbc instanceof Nodes) && !(wbc instanceof Voltmeter)
                && !(wbc instanceof Ammeter))
            {
 //               circuit.removeNodes(t[0].getID(), t[1].getID());
         	Point locwbc= ((CircuitComponent)wbc).getLocation();
                Point locterm = new Point ((t[0].getBounds()).x,(t[0].getBounds()).y);
                 
                Point loc= new Point (locwbc.x+locterm.x,locwbc.y+locterm.y);	
            	N1 =new Nodes(t[0]);
           	N1.setWorkbench(wb);
                N1.setLocation(loc);
            	wb.add(N1,0);
            	wb.componentAdded(N1,null); 
            }
            else
            {  // remove the links to this terminal
                circuit.removeNodes(t[0].getID(), t[1].getID());
                t[0].destroyLinks(wb);
            } 
            if( t[1].numLinks() >1 && !(wbc instanceof Nodes) && !(wbc instanceof Voltmeter)
                && !(wbc instanceof Ammeter))
            {
//               circuit.removeNodes(t[0].getID(), t[1].getID());
         	Point locwbc= ((CircuitComponent)wbc).getLocation();
                Point locterm = new Point ((t[1].getBounds()).x,(t[1].getBounds()).y);
                Point loc= new Point (locwbc.x+locterm.x,locwbc.y+locterm.y);	
            	N2 =new Nodes(t[1]);
            	N2.setWorkbench(wb);
                N2.setLocation(loc);
            	wb.add(N2,0);
            	wb.componentAdded(N2,null); 
            } 
            else
            {  // remove the links to this terminal
               t[1].destroyLinks(wb);
            } 
            wb.validate();
//	    circuit.removeNodes(t[0].getID(), t[1].getID());
	    // remove component?
	    circuit.findLoops();
	}
    }

    public void componentEdited(WorkbenchComponentEvent e) {
	WorkbenchComponent wbc = e.getWorkbenchComponent();
	// Make sure this is a CircuitComponent
	if (wbc instanceof CircuitComponent) {
	    circuit.findLoops();
	}
    }

}





