package sieve.sluice;

import java.net.URL;

/**
 * Abstract class representing a simple Table of data. This definition
 * makes several simplifying assumptions: 
 * <UL>
 * <LI> The table is rectangular 
 * <LI> The table is completely filled
 * <LI> Fields may <I>not</I> be added if there is data in the table
 * </UL>
 * If these assumptions are too restrictive for a particular application,
 * note that it is <I>not</I> necessary for a class to extend this class
 * in order to act as a data source -- a source need only implement the
 * <CODE>TableView</CODE> interface to act as a source for other 
 * <CODE>TableView</CODE>s.
 */
public abstract class Table implements TableView, TableEditor {

    protected TableViewEventSupport	eventHandler	= 
	new TableViewEventSupport();

    /* ---------------  Table methods --------------- */

    public abstract void insert(Object[][] data, int atRecord) 
	throws InterchangeException;

    public void append(Object[][] data)
	throws InterchangeException {
	insert(data, getRecordCount());
    }

    public abstract void addField(FieldDescriptor fd)
	throws InterchangeException;

    public abstract void remove(int firstRecord, int lastRecord)
        throws InterchangeException;

    public abstract void removeField(String fieldName) 
	throws InterchangeException;

    // These guys should be deprecated? Probably not: we still want these
    // to be settable as a property. Maybe we need a URL property editor.
    public String getLocation() {
	String loc = null;
        if (getLocationURL() != null)
            loc = getLocationURL().toString();
        return loc;
    }
    public void setLocation(String location) throws InterchangeException {
	URL newURL = URLFinder.getURLFinder().getURL(location);
        if (newURL != null)
            setLocationURL(newURL);
    }

    public abstract URL getLocationURL();
    public abstract void setLocationURL(URL location)
	throws InterchangeException;

	
    /* -------------  TableView methods ------------- */
    public TableDescriptor getTableDescriptor() throws NoSourceException {
        return new TableDescriptor();
    }

    public TableView table() {
	return this;
    }

    public void addTableViewListener(TableViewListener listener) {
        eventHandler.addTableViewListener(listener);
    }
 
    public void removeTableViewListener(TableViewListener listener) {
        eventHandler.removeTableViewListener(listener);
    }

    /**
     * For now we will just leave addSourceView and removeSourceView
     * as noops. Should we throw an exception instead?
     */
    public void addSourceView(TableView source) 
	throws TooManySourcesException {
    }
    public void removeSourceView(TableView source) { }

    /**
     * Propagates event to all listeners with no filtering.
     * This implemented primarily for compliance with the TableView 
     * interfaces, since Tables will typically not have source TableViews
     * and will therefore typically not receive events.
     */
    public void tableViewChanged(TableViewEvent event) {
        eventHandler.fireTableViewEvent(event);
    }

}
