package sieve.sluice.codebook;

import java.net.*;
import java.util.*;
import java.io.*;
import sieve.sluice.*;

/**
 * A CodataTable reads codebook files in the CODATA format.  This is one 
 * of the formats that the 1990 census data comes in on the web.
 */

public class CodataTable extends CodeBookTable {
    
    public CodataTable(String url) throws InterchangeException {
        super(url);
    }
    
    public CodataTable(URL url) throws InterchangeException {
        super(url);
    }
    
    public CodataTable() throws InterchangeException {
        super("http://simon.cs.vt.edu/~jnielsen/data/census.cod");
    }

    /**
     * Returns a name and value pair from the string.  Everything up to the
     * equal sign (trimmed) is the name, and everything following is the value.
     * Quotes and pound signs are stripped from the value.
     */
    private NVPair getNVPair(String s) {
        s = s.replace('\"', ' ');
        s = s.replace('#', ' ');
        int eqIndex = s.indexOf('=');
        if (eqIndex > 0) {
            String name = s.substring(0, eqIndex).trim();
            String val = s.substring(eqIndex + 1).trim();
            return new NVPair(name, val);
        }
        else return null;       // Signals a comment.
    }

    
    /**
     * Parses the codebook file for all of the required information.
     * Sets the following class variables:<UL>
     * <LI> dataFile
     * <LI> rowsAreFields
     * <LI> dataDelim
     * <LI> missingVal
     * <LI> attributes
     * <LI> fields
     * <LI> fieldIndices
     * </UL>
     *
     * This is the only important method that is overridden in this
     * subclass.
     */
    protected void parseCodebookFile() throws IOException, MalformedURLException,
        InterchangeException, SyntaxException {

        //InputStream is = codebookFile.startsWith("http://")
        //                 ? (new URL(codebookFile)).openStream()
        //                 : new FileInputStream(codebookFile);
	URL book = getLocationURL();
        if (book == null)
            return;
        InputStream is = book.openStream();


        // Open streams
        LineNumberInputStream cbStream = 
            new LineNumberInputStream(new BufferedInputStream(is));
        DataInputStream dcbStream = new DataInputStream(cbStream);

        // Set defaults for this type of codebook
        rowsAreFields = false;  // Can be changed with the "ROW=" attribute
        dataDelim = "\t";       // For tab-delimited census data--can be changed
                                // with "DELIM=" attribute
                                //
        missingVal = ".";       // WARNING: Census data default is to use empty
                                // string for missing value.  But empty string
                                // doesn't parse correctly.

        // Set up structures to hold variable information
        VariableVector tempFields = new VariableVector();
        fieldIndices = new Hashtable();
        int varCount = 0;
        Variable currVar = null;


        boolean eof = false;
        // Now read all of the name-value pairs from the file in order
loop:   while (! eof) {  // Exit loop with break
            NVPair nv;
            do { 
                String line = dcbStream.readLine();
                if (line == null) {
                    eof = true;
                    break loop;     // LOOP EXIT ON EOF
                }
                nv = getNVPair(line); 
            } while (nv == null);

            if (nv.name.equals("DATAFILE")) {
                dataFile = nv.val;
            }
            else if (nv.name.equals("ROW")) {
                if (! nv.val.equalsIgnoreCase("RECORD")) rowsAreFields = true;
            }
            else if (nv.name.equals("DELIM")) {
                dataDelim = nv.val;
            }
            else if (nv.name.equals("DE")) {        // New field
                currVar = new Variable(nv.val);
                tempFields.addElement(currVar);
                fieldIndices.put(currVar.getName(), new Integer(varCount++));
            }
            
            else if (nv.name.equals("TYPE")) {
                if (currVar == null)
                    throw new InterchangeException("No current variable.");
                // I don't know all of the possible CODATA types.
                // Right now just assign type "A" to a String and type
                // "I" to an Integer
                if (nv.val.equals("I")) 
                    currVar.setType("java.lang.Integer");
                else currVar.setType("java.lang.String");
            }
            else if (nv.name.equals("USE")) {
                if (currVar == null)
                    throw new InterchangeException("No current variable.");
                // I don't know what this means.
            }
            else if (nv.name.equals("LENGTH")) {
                if (currVar == null)
                    throw new InterchangeException("No current variable.");
                if ((nv.numberVal() > 9) && 
                    currVar.getType().equals("java.lang.Integer"))
                    currVar.setType("java.lang.Long");
            }
            else if (nv.name.equals("HEADER")) {
                if (currVar == null)
                    throw new InterchangeException("No current variable.");
                if (! nv.val.startsWith("STF")) {
                    String oldDesc = currVar.getDescription();
                    if (oldDesc.equals(currVar.getName()))
                        currVar.setDescription(nv.val);
                    else currVar.setDescription(oldDesc  + "/" + nv.val);
                }
            }

        }

        // Now we have all of the Variables--transfer them to fields array
        fields = tempFields.getArray();
        // Set levels.  Assume that ints are ratio data, strings are nominal
        for (int i = 0; i < fields.length; i++) {
            if (fields[i].getType() == "java.lang.Integer") 
                fields[i].setLevel(FieldDescriptor.LEVEL_RATIO);
            else if (fields[i].getType() == "java.lang.String")
                fields[i].setLevel(FieldDescriptor.LEVEL_NOMINAL);
        }

        // Close codebook file
        cbStream = null;
    }

}

class NVPair {
    String name;
    String val;

    NVPair(String n, String v) { name = n; val = v; }

    int numberVal() {
        return Integer.parseInt(val);
    }
}






