package sieve.sluice.codebook;

import java.net.*;
import java.util.*;
import java.io.*;
import sieve.sluice.*;

public class SPSSTable extends CodeBookTable {
    
    public SPSSTable(String url) throws InterchangeException {
        super(url);
    }

    public SPSSTable(URL url) throws InterchangeException {
        super(url);
    }
    
    
    public SPSSTable() throws InterchangeException {
        super("http://simon.cs.vt.edu/~jnielsen/data/all93.sps");
    }

    /**
     * Returns the longest quoted string contained in the string
     */
    private String extractQuoted(String s) {
        String temp = s.replace('\'', '\"');
        return temp.substring(temp.indexOf('\"') + 1, temp.lastIndexOf('\"'));
    }

    
    /**
     * Parses the codebook file for all of the required information.
     * Sets the following class variables:<UL>
     * <LI> dataFile
     * <LI> rowsAreFields
     * <LI> dataDelim
     * <LI> missingVal
     * <LI> attributes
     * <LI> fields
     * <LI> fieldIndices
     * </UL>
     *
     * This is the only important method that is overridden in this
     * subclass.
     */
    protected void parseCodebookFile() throws IOException, MalformedURLException,
        InterchangeException, SyntaxException {

        //InputStream is = codebookFile.startsWith("http://")
        //                 ? (new URL(codebookFile)).openStream()
        //                 : new FileInputStream(codebookFile);

	URL book = getLocationURL();
        if (book == null)
            return;
        InputStream is = book.openStream();


        LineNumberInputStream cbStream = 
            new LineNumberInputStream(new BufferedInputStream(is));

        DataInputStream dcbStream = new DataInputStream(cbStream);

        // Get dataFile name from DATA LIST statement
        String line = dcbStream.readLine();
        while (! line.startsWith("DATA LIST FREE FILE=")) line = dcbStream.readLine();
        dataFile = extractQuoted(line);
//        System.out.println("Got dataFile name: " + extractQuoted(line));


        // Set rowsAreFields--false by default unless someone has
        // added TRANSPOSE to end of line
        rowsAreFields = (line.endsWith("TRANSPOSE") ? true : false);
        // set dataDelim & missingVal--always a comma and period with these files
        dataDelim = ",";
        missingVal = ".";

        // No attributes to put
    
        // Now read variable information from rest of file
        VariableVector tempFields = new VariableVector();
        fieldIndices = new Hashtable();
        int varCount = 0;
        Variable currVar = null;

        do { line = dcbStream.readLine(); } while (! line.startsWith("/"));
        for (line = dcbStream.readLine(); ! line.startsWith("."); 
            line = dcbStream.readLine()) {
                
            currVar = new Variable(line.trim());
            tempFields.addElement(currVar);
            fieldIndices.put(currVar.getName(), new Integer(varCount++));

//            System.out.println("Got variable name: " + currVar.getName());

        }
        // Now we have all of the Variables
        fields = tempFields.getArray();

        // Now get other information
        boolean finished = false;
        while (! finished) {
            int i = 0;
            String section = dcbStream.readLine();
            if (section == null) finished = true;
            else if (section.startsWith("VARIABLE")) {
                // Get the description of each variable--assume each has one
                for (line = dcbStream.readLine(); ! line.startsWith("."); 
                    line = dcbStream.readLine()) {
                    // Should really match w/name--there could be missing descriptions
                    fields[i++].setDescription(extractQuoted(line));
//                    System.out.println(extractQuoted(line));
                }
            }
            else if (section.startsWith("VALUE")) {
                // Record the value labels for each variable
                for (line = dcbStream.readLine(); ! line.startsWith("."); ) {
                    Variable v;
                    do {        // Match with correct variable--assume same order
                        v = fields[i++];
                    } while (! v.getName().equals(line.replace('/', ' ').trim()));
                    while (true) try {
                        line = dcbStream.readLine();
                        Integer val = new Integer(
                            line.substring(0, line.indexOf('\"')).trim());
                        v.addValLabel(val, extractQuoted(line));
                    } catch (NumberFormatException e) {
                        Float val = new Float(
                            line.substring(0, line.indexOf('\"')).trim());
                        v.addValLabel(val, extractQuoted(line));
                    } catch (StringIndexOutOfBoundsException e) {
                        break;  // To next iteration of for
                    }
                }
            }
            else if (section.startsWith("FORMAT")) {
                for (line = dcbStream.readLine(); ! line.startsWith("."); 
                    line = dcbStream.readLine()) {
                    // This is a hack--I'm just looking too see if there's a 
                    // zero after the decimal point.  If there is, it's an int.
                    // Otherwise it's a float.  Should try to figure out
                    // what the SPSS format representations mean & translate them.
                    if (line.charAt(line.indexOf('.') + 1) == '0')
                        fields[i++].setType("java.lang.Integer");
                    else fields[i++].setType("java.lang.Float");
                }
            }
            else if (section.startsWith("MISSING")) {
                for (line = dcbStream.readLine(); ! line.startsWith("."); 
                    line = dcbStream.readLine()) {
                    // Don't know what to do with these
                }
            }
            else finished = true;   // There shouldn't be other sections
        }

        // Close codebook file
        cbStream = null;
    }

}








