/**
 * 
 * STFTable.java
 * 
 * This class allows TableView access to STF (Simple Tabulated Format) files.
 * It extends CodeBookTable to get a good chunk of it's code, but because
 * of the nature of the standard codebooktable, it overwrites a great many
 * of the methods therein.
 *
 * Christopher Telfer :: ctelfer@vt.edu 
 * 4/3/98
 *
 */

package sieve.sluice.codebook; 

import java.net.*; 
import java.util.*; 
import java.io.*; 
import sieve.sluice.*; 
import java.util.*;

public class STFTable extends CodeBookTable
{ 

  private static final String delims = " \t";
  private BufferedReader data = null; 

  // This is really cheating, but it makes it easy to skip past the header
  // when we have to open the file for actual data fetching.
  private int linesInHeader = 0; 

  // Constructors
  public STFTable(String url) throws InterchangeException { 
     setLocation(url); 
  }

  public STFTable(URL url) throws InterchangeException { 
     setLocationURL(url); 
  }

  // needs to be overwritten
  public void setLocationURL(URL codebookURL) throws InterchangeException
  { 
    try { 
      this.codebookURL = codebookURL; 
      System.out.println("Parsing the codebook header."); 
      // this should set the variable reader for our usage later!
      parseCodebookFile(); 
    }
    catch(Exception e) { 
      e.printStackTrace(); 
      throw new InterchangeException
                ("Error reading the codebook header" + codebookURL); 
    }

    // From here, now that we have our fields and what not, we can look 
    // at the data for purposes of getting the number of records and stuff.

    System.out.println("Parsing the codebook data."); 
    try 
    { 
      int numFields = fields.length; 
      String line = null; 
      do 
      { 
        line = data.readLine(); 
	if ((line != null)  && (line.charAt(0) != '#')) numRecords++;
      } while (line != null); 

      data.close(); 
    }
    catch(Exception e) { 
      e.printStackTrace(); 
      throw new InterchangeException
                ("Error reading/parsing the data section of the codebook"); 
    }


  }

  /**
   *
   * Parses the codebook header for all information.
   * Sets the following variables : 
   * <UL>
   * <LI> fields
   * <LI> fieldIndices
   * </UL>
   *
   */
  protected void parseCodebookFile() throws IOException, MalformedURLException,
                 InterchangeException, SyntaxException, NumberFormatException
  { 
    int numFields = 0, i; 

    if (codebookURL == null) return; 
    InputStream is = codebookURL.openStream(); 

    data = new BufferedReader(new InputStreamReader(is)); 

    // From here what we do is parse the header
    String line = data.readLine(); 
    while ((line != null) && (line.charAt(0) == '#'))
    { 
      linesInHeader++; 
      line = data.readLine(); 
    }

    if (line == null) throw new InterchangeException("Incomplete Header"); 
    else linesInHeader++; 


    numFields = java.lang.Integer.parseInt(line); 
    fieldIndices = new Hashtable(); 
    fields = new Variable[numFields]; 
    i=0; 
    while (i < numFields)
    { 
       line = data.readLine(); 
       if (line == null) throw new InterchangeException("Incomplete Header"); 
       linesInHeader++; 
       if   (line.charAt(0) == '#') continue; 

       fields[i] = new Variable(line); 
       fieldIndices.put(line, new Integer(i)); 
       i++; 
    }
    
  }

  public FieldDescriptor getFieldDescriptor(String field) throws 
         InterchangeException 
  {
    Variable fld = fields[fieldIndex(field)];
    fld.setLevel(FieldDescriptor.LEVEL_NOMINAL);
    return fld; 
  }

  public Object[][] get(String flds[], int startingAt, int numRecs)
         throws InterchangeException
  { 
    String line = null; 
    Object[][] retArray = new Object[flds.length][numRecs]; 
    String[] fieldsInLine = new String[fields.length]; 
    int numRead = 0, readPast = startingAt; 

    if ((startingAt + numRecs > numRecords) || (startingAt < 0))
        throw new InterchangeException("Request out of bounds"); 
    for (int i = 0;i<flds.length;i++)
        if (fieldIndices.get(flds[i]) == null)
	  throw new sieve.sluice.NoSuchFieldException(flds[i]); 
    try
    { 
      data= new BufferedReader(new InputStreamReader(codebookURL.openStream()));
      for(int i=0;i<linesInHeader;i++)
         line = data.readLine(); 

      while (readPast > 0)
      { 
        line = data.readLine(); 
	if (line == null) throw new InterchangeException(); 
	if (line.charAt(0) == '#') continue; 
	readPast--; 
      }

      while (numRead < numRecs)
      {
        line = data.readLine(); 
	if (line == null) throw new InterchangeException(); 
	if (line.charAt(0) == '#') continue; 

	StringTokenizer st = new StringTokenizer(line, delims);
	for(int i = 0;i < fields.length; i++)
	{ 
	  if (!st.hasMoreTokens()) 
	    throw new InterchangeException("Not enough tokenson data line " + 
	              (numRead + readPast)); 
		       
          fieldsInLine[i] = st.nextToken(); 
	}

	for(int i = 0;i < flds.length;i++)
	   retArray[i][numRead] = fieldsInLine[fieldIndex(flds[i])]; 

	numRead++; 
      } // while reading in records

    } // try block
    catch(Exception e)
    { 
      e.printStackTrace(); 
      if (e instanceof InterchangeException)
         throw (InterchangeException)e; 
      else throw new InterchangeException("Error reading site."); 
    }

    return retArray; 
     
  }

  private int fieldIndex(String fieldName) throws InterchangeException
  { 
     Object intObj = fieldIndices.get(fieldName); 
     if (intObj == null) 
        throw new sieve.sluice.NoSuchFieldException(fieldName); 
     return ((Integer)intObj).intValue(); 
  }

  public TableDescriptor getTableDescriptor()
  { 
    return new TableDescriptor(); 
  }


} // STFTable
