package sieve.sluice.views;

import sieve.sluice.*;
import java.util.*;

/**
 * A DescriptiveView is a view of all values of a single field with
 * methods for determining frequencies and descriptive statistics
 * commonly needed by visualizations.
 */
public class DescriptiveView extends FieldNameSelectionView {
    /**
     * Get BLOCK_SIZE records at a time when initializing.
     */
    protected	static final int BLOCK_SIZE = 2000;
    protected 	String 		field		= null;
    protected   boolean		isNominal	= false;
    protected   boolean		isOrdinal       = false;
    protected	boolean		isNumeric	= false;

    protected	Hashtable	values		= null;
    protected	Vector		sortedValues	= null;

    protected	int		maxFrequency	= 0;
    protected	Vector		mode		= null;

    protected	double		sum		= 0.0;

    /**
     * true if variance need to be recalculated.
     */
    protected	boolean		varianceIsDirty		= true;
    protected	double		variance			= 0.0;

    public DescriptiveView(TableView source, String field) throws 
	InterchangeException {
	super(source, field);
	this.field = field;
	reset();
    }

    /**
     * Destroys the view's current represenation of the field's contents
     * and rebuilds it from scratch.
     */
    protected synchronized void reset() throws InterchangeException {
	values = new Hashtable();
	sortedValues = new Vector();
	mode = new Vector();

	int records = getRecordCount();

	if (records == 0) {
	    return;
	}

        isNominal = (getFieldDescriptor(field).getLevel() == 
	    FieldDescriptor.LEVEL_NOMINAL);

        isOrdinal = (getFieldDescriptor(field).getLevel() == 
	    FieldDescriptor.LEVEL_ORDINAL);

	// This gets reset below.
	isNumeric = false;

	int recsToGet = BLOCK_SIZE;
	Object[] data;
	for (int i = 0; i < records; ) {
	    recsToGet = records - i;
	    if (recsToGet > BLOCK_SIZE)
		recsToGet = BLOCK_SIZE;

	    data  = get(i, recsToGet);
	    for (int j = 0; j < recsToGet; j++) {
		/*
		 * Need to think about this a bit more. Handling of
		 * missing values should do different things for 
		 * numbers...
		 */
/* FIX THIS...
		if (data[j] == null)
		    addValue("(Unknown)");
		else
		    addValue(data[j]);
*/
		if (data[j] != null) {
		    isNumeric = isNumeric || (data[j] instanceof Number);
		    addValue(data[j]);
		}

		i++;
	    }
	}
    }


    /**
     * Increments the view's counter of the number of occurances
     * of the specified value. Need to handle missing (null) values here.
     */
    protected synchronized void addValue(Object value) {
	ValueCounter counter = (ValueCounter) values.get(value);
	if (counter == null) {
	    counter = new ValueCounter();
	    values.put(value, counter);
	    if (isNumeric) {
		// Make this a binary search...
		int k = 0;
		for (k = 0; (k < sortedValues.size()) &&
		    ((Number)(sortedValues.elementAt(k))).doubleValue()
		    < ((Number)(value)).doubleValue(); k++);
		    sortedValues.insertElementAt(value, k);
	    }
	    else 
		sortedValues.addElement(value);
	}
	counter.count++;

	if (isNumeric) {
	    sum += ((Number)value).doubleValue();
	}

	if (counter.count == maxFrequency) {
	    mode.addElement(value);
	}
	else if (counter.count > maxFrequency) {
	    maxFrequency = counter.count;
	    mode.removeAllElements();
	    mode.addElement(value);
	}
	varianceIsDirty = true;
    }

    /**
     * Decrements the view's counter of the number of occurances
     * of the specified value.
     */
    protected synchronized void removeValue(Object value) {
	ValueCounter counter = (ValueCounter) values.get(value);
	if (counter != null) {
	    counter.count--;
	    if (counter.count == 0) {
		values.remove(value);
		sortedValues.removeElement(value);
	    }
	    if (mode.contains(value)) {
		mode.removeElement(value);
		if (mode.size() == 0) {
		    maxFrequency = 0;
		    for (Enumeration e = sortedValues.elements(); 
		        e.hasMoreElements();) {
		        Object obj = e.nextElement();
		        ValueCounter c = (ValueCounter) values.get(obj);
		        if (c.count == maxFrequency) {
			    mode.addElement(obj);
			}
		        if (c.count > maxFrequency) {
			    maxFrequency = c.count;
			    mode.removeAllElements();
			    mode.addElement(obj);
		        }
		    }
		}
	    }
	    if (isNumeric) 
		sum -= ((Number)value).doubleValue();
        }
	varianceIsDirty = true;
    }

    /**
     * Convenience method for getting raw data.
     */
    public synchronized Object[] get(int startingAt, int numberOfRecords) 
	throws InterchangeException {
	Object[][] data = get(getFieldNames(), startingAt, numberOfRecords); 
	if (data == null)
	    return null;
	else
	    return data[0];
    }

    /**
     * Convenience method for getting raw data.
     */
    public synchronized Object[] get(int startingAt) throws 
	InterchangeException {
	return get(startingAt, getRecordCount() - startingAt);
    }

    /**
     * Returns the minimum value of the data if it is not nominal.
     */
    public synchronized Object getMin() throws WrongDataKindException {
	try {
	    if (isNumeric && ! isNominal) 
		return sortedValues.elementAt(0);
	}
	catch (Exception e) {}
	throw new WrongDataKindException();
    }
	
    /**
     * Returns the maximum value of the data if it is not nominal.
     */
    public synchronized Object getMax() throws WrongDataKindException {
	try {
	    if (isNumeric && ! isNominal) 
		return sortedValues.elementAt(sortedValues.size() - 1);
	}
	catch (Exception e) { }
	throw new WrongDataKindException();
    }
	
    /**
     * Returns the mean of the data if it is numeric and not nominal
     * or ordinal.
     */
    public synchronized double getMean() throws WrongDataKindException {
	try {
	    if (isNumeric && ! isNominal && ! isOrdinal)  
		return sum/getRecordCount();
	}
	catch (Exception e) { }
	throw new WrongDataKindException();
    }

    /**
     * Returns the variance of the data if it is numeric and non-nominal.
     * Division by n-1 (rather than n) is used, giving an unbiased
     * estimate of the population variance (use getVarianceMLE()
     * for a maximum likelihood estimate.
     */
    public synchronized double getVariance() throws WrongDataKindException {
	if (!varianceIsDirty) {
	    return variance;
	}
	else {
	    try {
		if (isNumeric && ! (isNominal || isOrdinal)) {
		    double mean = getMean();
		    double varSum = 0.0;
		    for (Enumeration e = sortedValues.elements(); 
			e.hasMoreElements();) {
			Number n = (Number) e.nextElement();
			ValueCounter counter = (ValueCounter) values.get(n);
			double diff = n.doubleValue() - mean;
			varSum += counter.count * diff * diff;
		     }
		     variance = varSum/(getRecordCount() - 1);
		     varianceIsDirty = false;
		     return variance;
		}
	    }
	    catch (Exception e) { }
	    throw new WrongDataKindException();
	}
    }

    /**
     * Returns the variance of the data if it is numeric and non-nominal.
     * Division by n (rather than n-1) is used, giving a maximum
     * likelihood estimate of the population variance (use getVariance()
     * for an unbiased estimate).
     */
    public synchronized double getVarianceMLE() throws InterchangeException {
	return (getVariance() * (getRecordCount() - 1))/getRecordCount();
    }

    /**
     * Returns the standard deviation of the data if it is 
     * numeric and non-nominal.
     * Division by n-1 is used, giving an unbiased estimate of the
     * population standard deviation.
     */
    public synchronized double getStandardDeviation() 
	throws WrongDataKindException {
	return Math.sqrt(getVariance());
    }

    /**
     * Returns the standard deviation of the data if it is 
     * numeric and non-nominal.
     * Division by n is used, giving a maximum likelihood estimate of
     * the population standard deviation.
     */
    public synchronized double getStandardDeviationMLE() 
	throws InterchangeException {
	return Math.sqrt(getVarianceMLE());
    }


    /**
     * Returns the median (middle value) of the data if it is 
     * non-nominal.
     */
    public synchronized Object getMedian() throws WrongDataKindException {
	try {
	    if (! isNominal) {
		int position = getRecordCount()/2;
		int current  = 0;
		for (Enumeration e = sortedValues.elements(); 
		    e.hasMoreElements();) {
		    Object obj = e.nextElement();
		    ValueCounter counter = (ValueCounter) values.get(obj);
		    current += counter.count;
		    if (current >= position)
		         return obj;
		}
	    }
	}
	catch (Exception e) { }
	throw new WrongDataKindException();
    }

    /**
     * Returns the mode (most common value) of the data.
     */
    public synchronized Object[] getMode() {
	Object[] modes = new Object[mode.size()];
	mode.copyInto(modes);
	return modes;
    }

    /**
     * Returns the number of occurances of the given value. We do
     * not throw an exception if the value isn't found, since it may
     * have been valid but simply not found in this table.
     */
    public synchronized int getFrequency(Object value) {
	ValueCounter counter = (ValueCounter)(values.get(value));
	if (counter == null)
		return 0;
	else
		return counter.count;
    }

    /**
     * Returns the number of occurances of values between the given values
     * <I>Inclusive</I>.
     */
    public synchronized int getFrequency(Number min, Number max) 
	throws WrongDataKindException{
	int count = 0;
	try {
	    if (isNumeric && !isNominal) { 
		Enumeration e = sortedValues.elements();
		Number n = null;
		while (e.hasMoreElements() && 
		    ((n = (Number)(e.nextElement())).doubleValue() < 
		    min.doubleValue()));
		while ((n != null) && (n.doubleValue() <= max.doubleValue())) {
		    ValueCounter counter = (ValueCounter)values.get(n);
		    count += counter.count;
		    if (e.hasMoreElements())
			n = (Number)(e.nextElement());
		    else
			n = null;
		}
	    }
	    return count;
	}
	catch (Exception ex) {}
	throw new WrongDataKindException();
    }

    /**
     * Returns a vector of all distinct values found in
     * the data. If the data in not nominal, the vector will be ordered
     * from minimum value to maximum value.
     */
    public Vector getValues() {
	return (Vector) sortedValues.clone();
    }

    public void tableViewChanged(TableViewEvent event) {
	/*
	 * For now, just blast everything...
	 */
	try {
	    reset();
	}
	catch (Exception e) {
	     e.printStackTrace();
	}
        super.tableViewChanged(event);
    }


    class ValueCounter {
        public int count = 0;
        public ValueCounter() {
	    count = 0;
        }
    }
}


