package sieve.sluice.views;

import sieve.sluice.*;
import sieve.sluice.event.*;
import sieve.sluice.NoSuchFieldException.*;
import java.util.*;
import java.awt.*;
import java.beans.*;

/**
 * A FrequencyTableView generates a new two-field table containing the
 * values and frequencies of 
 * the contents of a single field from the source table.
 */
public class FrequencyTableView implements TableView, TableViewListener  {
    protected final static String FREQ_FLD 		= "Frequency";
    protected final static String PERCENT_FLD 		= "Percent";
    protected final static String C_PERCENT_FLD 	= "Cumulative Percent";

    TableView			origSource	= null;
    Hashtable			fields		= null;
    String			sourceField	= null;
    FieldDescriptor		sourceFieldDescriptor = null;
    TableViewEventSupport	eventHandler	= null;
    PropertyChangeSupport       changes         = null;

    public FrequencyTableView() {
	this.eventHandler = new TableViewEventSupport();
	this.changes = new PropertyChangeSupport(this);
    }

    public FrequencyTableView(TableView source, String field) throws
	InterchangeException {
	this();
        this.sourceField = field;
	this.origSource = source;
	reset();
    }

    protected synchronized void reset() {
	if ((origSource != null) && (sourceField != null)) {
	    DescriptiveView source = null;
	    try {
		source = new DescriptiveView(origSource, sourceField);
	    }
	    // If our field disappeared, default back to the first
	    // available field
	    catch (sieve.sluice.NoSuchFieldException nsfe) {
		try {
		    String[] allFlds = origSource.getFieldNames();
		    if (allFlds.length > 0)
		        setField(allFlds[0]);
		    return;
		}
		catch(InterchangeException e) {
		    sourceField = null;
		    e.printStackTrace();
		    return;
		}
	    }
	    catch (InterchangeException e) {
		e.printStackTrace();
		return;
	    }
	    Vector valVector = source.getValues();
	    Object[] values = new Object[valVector.size()];
	    Long[] freqs = new Long[valVector.size()];
	    Double[] percents = new Double[valVector.size()];
	    Double[] c_percents = new Double[valVector.size()];
	    int i = 0;

	    for (Enumeration e = valVector.elements(); e.hasMoreElements();) {
		Object val = e.nextElement();
		values[i] = val;
		freqs[i] = new Long(source.getFrequency(val));
		i++;
	    }

	    try {
	        fields = new Hashtable();
                FieldDescriptor fd = source.getFieldDescriptor(sourceField);
	        fields.put(sourceField, new FreqFieldInfo(sourceField, 
		    fd.getType(), fd.getDescription(), 
		    fd.getLevel(), fd.getLabels(), values));

                sourceFieldDescriptor = fd;

		fields.put(FREQ_FLD, new FreqFieldInfo(FREQ_FLD, 
		    "java.lang.Long", null, FieldDescriptor.LEVEL_INTERVAL, 
                    null, freqs));
	    }
	    catch (InterchangeException e) {
		e.printStackTrace();
	    }
	}
    }

    /**
     * Returns a reference to this view. This is necessary since the
     * structure of this TableView is radically different from that of
     * our source TableView.
     */
    public TableView table() throws NoSourceException {
	if (origSource == null)
	    throw new NoSourceException();
	return this;
    }

    public TableDescriptor getTableDescriptor() {
      TableDescriptor td = new TableDescriptor();
      td.setLabelField(sourceField);
      return td;
    }

    public String[] getFieldNames() throws NoSourceException {
	if (origSource == null)
	    throw new NoSourceException();
	//String[] flds = {sourceField, FREQ_FLD, PERCENT_FLD, C_PERCENT_FLD};
	String[] flds = {sourceField, FREQ_FLD}; 
	return flds;
    }

    public FieldDescriptor getFieldDescriptor(String field) throws 
        InterchangeException {
	if (origSource == null)
	    throw new NoSourceException();
	FreqFieldInfo info = (FreqFieldInfo) fields.get(field);
	if (info == null)
	    throw new sieve.sluice.NoSuchFieldException(field);
        FieldDescriptor fd = new FieldDescriptor(field);
        if (field.equals(sourceField))
            fd.setColor(Color.black);
        else
            fd.setColor(sourceFieldDescriptor.getColor());
        fd.setType(info.type);
        fd.setDescription(info.description);
        fd.setLabels(info.labels);
        fd.setLevel(info.kind); // WRONG!!
	return fd;
    }

    /**
     * Returns number of records in the view.
     */
    public synchronized int getRecordCount() throws NoSourceException {
	if ((origSource == null) || (sourceField == null) || (fields == null))
	    throw new NoSourceException();
	FreqFieldInfo info = (FreqFieldInfo) fields.get(sourceField);
	if (info == null)
	    return 0;
	else
	    return info.values.length;
    }

    public synchronized Object[][] get(String[] flds, int startingAt, 
	int numberOfRecords) throws InterchangeException {
	if (origSource == null)
	    throw new NoSourceException();
	if (startingAt >= getRecordCount()) 
		throw new sieve.sluice.NoSuchRecordException(startingAt);
	if (startingAt + numberOfRecords > getRecordCount())
		throw new sieve.sluice.NoSuchRecordException(
		    getRecordCount() + 1);

	Object[][] data = 
	     new Object[flds.length][(int)(numberOfRecords - startingAt)];
	FreqFieldInfo[] fldRecs = new FreqFieldInfo[flds.length];
	for (int i = 0; i < flds.length; i++) {
	    FreqFieldInfo info = (FreqFieldInfo)fields.get(flds[i]);
	    if (info == null)
		throw new sieve.sluice.NoSuchFieldException(flds[i]);
	    else
		fldRecs[i] = info;
	}
	for (int fld = 0; fld < flds.length; fld++) {
	    for (int rec = 0; rec < numberOfRecords; rec++) {
		data[fld][rec] = fldRecs[fld].values[(int)(rec + startingAt)];
	    }
	}
	return data;
    }

    /**
     * Returns all records starting from record number <CODE>startingAt</CODE>
     * from the fields specified in <CODE>fields</CODE>.
     */
    public synchronized Object[][] get(String[] flds, int startingAt) 
	throws InterchangeException { 
	if (origSource == null)
	    throw new NoSourceException();
	return get(flds, startingAt, getRecordCount()); 
    } 

    /* -------------  Event-related methods ------------- */
    public synchronized void addTableViewListener(TableViewListener listener) {
	eventHandler.addTableViewListener(listener);
    }

    public synchronized void 
	removeTableViewListener(TableViewListener listener) {
	eventHandler.removeTableViewListener(listener);
    }


    /* -------------  Methods for link support ------------- */
    /**
     * By default we throw an exception if we already have a source.
     * If we don't, then an event is sent to all of our listeners.
     * We DO NOT add ourselves as a listener to the new source. This
     * is the responsibility of whatever is connecting us.
     */
    public synchronized void addSourceView(TableView source) 
        throws TooManySourcesException {
        if (this.origSource != null)
	    throw new TooManySourcesException("Only one source allowed.");

	this.origSource = source;
        if (sourceField != null) 
	    setField(sourceField);
	eventHandler.fireTableViewEvent(new SourceAddedEvent(this));
    }

    public synchronized void setField(String field) {
	// Need to revisit this check sometime.
	//if ((field != null) && ((this.sourceField == null) || 
	//    (! field.equals(this.sourceField)))) {
	    String oldField = this.sourceField;
	    this.sourceField = field;
	    reset();
	    eventHandler.fireTableViewEvent(new TableViewEvent(this));
	    changes.firePropertyChange("field", oldField, field);
	//}
    }

    public synchronized String getField() {
	return sourceField;
    }

    /**
     * If the parameter is actually our current source, we remove it
     * and fire an event to anyone listening to us.
     */
    public synchronized void removeSourceView(TableView source) {
        if (source == this.origSource) {
            this.origSource = null;
	    eventHandler.fireTableViewEvent(new SourceRemovedEvent(this));
	}
    }

    public synchronized void tableViewChanged(TableViewEvent event) {
	// If the field descriptor changed, this could affect downstream
	// modules' representations of both the source and frequency fields.
	// We therefore fire an additional event.
	if (event instanceof FieldDescriptorChangeEvent) {
	    String changedField = 
		((FieldDescriptorChangeEvent)event).getField();
	    if (changedField.equals(sourceField)) {
		try {
		    sourceFieldDescriptor = 
		        origSource.getFieldDescriptor(sourceField);
		}
		catch (InterchangeException e) {
		    return;
		}
		TableViewEvent freqEvent = 
		    new FieldDescriptorChangeEvent(this, FREQ_FLD);
		eventHandler.fireTableViewEvent(freqEvent);
	    }
	    // Stop propagation if this change affects a field we don't
	    // pass along.
	    else if (!changedField.equals(FREQ_FLD))
		return;
	}
	else
	    reset();
	eventHandler.fireTableViewEvent(event);
    }

    //----------------------------------------------------------------------
    // Methods for registering listeners:
    /**
     * The specified PropertyChangeListeners <b>propertyChange</b> method will
     * be called each time the value of any bound property is changed.
     *
     * @see #removePropertyChangeListener
     * @param l the PropertyChangeListener
     */
    public void addPropertyChangeListener(PropertyChangeListener l) {
        changes.addPropertyChangeListener(l);
    }
 
    /**
     * Remove this PropertyChangeListener from the JellyBeans internal list.
     * If the PropertyChangeListener isn't on the list, silently do nothing.
     *
     * @see #addPropertyChangeListener
     * @param l the PropertyChangeListener
     */
    public void removePropertyChangeListener(PropertyChangeListener l) {
        changes.removePropertyChangeListener(l);
    }

}

class FreqFieldInfo {
    String 	name	= null;
    String 	type 	= null;
    String	description = null;
    int		kind	= 0;
    Hashtable 	labels 	= null;
    Object[]    values 	= null;

    public FreqFieldInfo(String name, String type, String description,
	int kind, Hashtable labels, Object[] values) {
	this.name = name;
	this.type = type;
	this.description = description;
	this.kind = kind;
	this.labels = labels;
	this.values = values;
    }
}
