package sieve.sluice.views;

import java.util.*;
import sieve.sluice.*;
import sieve.sluice.event.*;

/**
 * SortedView sorts the records presented by its source based on a 
 * specified key field. (Note that before a field is specified, the
 * records are not sorted.) 
 * Explicit support for sorting Numbers, Strings, and Dates is provided.
 * Other types are converted to Strings with the <CODE>toString</CODE>
 * and then sorted.
 * More flexible support for sorting other types should be added in
 * future.
 * <P>
 * The basic algorithm used is quicksort, with selection sort used
 * for short segments or when the quicksort algorithm gets "stuck", which
 * typically happens because of non-unique pivot selection.
 */
public class SortedView extends SampleView implements Sampler {
    private static final int IS_NUMBER 	= 0;
    private static final int IS_STRING 	= 1;
    private static final int IS_DATE 	= 2;
    private static final int IS_OTHER 	= 3;
   
    private int[] 	map 	  = new int[0];
    private String 	field 	  = null;
    private int 	fieldType = IS_OTHER;

    private int		numCompares = 0;
    private boolean	sortAscending = true;
    //private boolean	missingValuesAtEnd = true;

    public SortedView() {
	super();
	setSampler(this);
    }

    public void tableViewChanged(TableViewEvent evt) {
	if (! (evt instanceof FieldDescriptorChangeEvent)) {
	    if ((source != null) && (field != null))
	        try {
	            fieldType = getFieldType(field);
	        }
	        catch (InterchangeException e) {
		    map = new int[0];
		    field = null;
		    e.printStackTrace();
		    return;  // Bad field
	        }
	        startSort();
	}
	eventHandler.fireTableViewEvent(new TableViewEvent(this));
    }

    public void removeSourceView(TableView source) {
	super.removeSourceView(source);
	if (this.source == null) {
	    map = new int[0];
	    field = null;
	}
    }

    /**
     * Sets the field to be used as a key for sorting. 
     * For fields containing Numbers, Strings, or Dates, this will
     * have the expected results.
     * For all other types of fields, the toString() operator will
     * be applied to each object, and the records will be sorted
     * based on the results.
     */
    public synchronized void setKeyField(String field) {
	if ((source != null) && 
	    ((this.field == null) || (! field.equals(this.field)))) {
	    String oldField = this.field;
	    try {
	        fieldType = getFieldType(field);
	    }
	    catch (InterchangeException e) {
		e.printStackTrace();
		return;  // Bad field
	    }
	    this.field = field;
	    startSort();
	    eventHandler.fireTableViewEvent(new TableViewEvent(this));
	    changes.firePropertyChange("keyField", oldField, field);
	}
    }
    public synchronized String getKeyField() {
	return field;
    }

    /**
     * By default, values are sorted in ascending order. Setting this
     * property to false sorts values in descending order.
     */
    public synchronized void setSortAscending(boolean sortAscending) {
	if (sortAscending != this.sortAscending) {
	    if (map != null && map.length > 0) {
		for (int i = 0; i < map.length/2; i++) {
		    int tmp = map[i];
		    map[i] = map[map.length - i - 1];
		    map[map.length - i - 1] = tmp;
		}
	        eventHandler.fireTableViewEvent(new TableViewEvent(this));
	    }
	    changes.firePropertyChange("sortAscending", 
		new Boolean(!sortAscending), new Boolean(sortAscending));
	}
    }
    public boolean getSortAscending() {
	return this.sortAscending;
    }

/*
    public synchronized void setMissingValuesAtEnd(boolean missingValuesAtEnd) {
    }
    public boolean getMissingValuesAtEnd() {
	return this. missingValuesAtEnd;
    }
*/

    /**
     * C-ish compare function. Returns negative int if obj1 < obj2, 0 if equal,
     * positive int if obj1 > obj2. If the sortAscending property is
     * set to false, then these are reversed.
     */
    private int compare(Object obj1, Object obj2) {
	numCompares++;
	int result = 0;

	// Push missing values to the front? This should probably be
	// configurable...
	if ((obj1 == null) && (obj2 == null))
	    result = 0;
	else if (obj1 == null)
	    result = 1;
	else if (obj2 == null)
	    result = -1;

	else {
	  switch (fieldType) {
	    case IS_NUMBER:
		// This will bomb on very small doubles...
		result = (int)(((Number)obj1).doubleValue() - 
		    ((Number)obj2).doubleValue());
		break;
	    case IS_STRING:
		result = ((String)obj1).compareTo((String)obj2);
		break;
	    case IS_DATE:
		Date d1 = (Date)obj1;
		Date d2 = (Date)obj2;
		if (d1.equals(d2))
		    result = 0;
		else if (d1.before(d2))
		    result = -1;
		else
		    result = 1;
		break;
	    default:
	    case IS_OTHER:
		result = (obj1.toString()).compareTo(obj2.toString());
	  }
	}
	if (!sortAscending)
	    result = - result;
	return result;
    }

    private int getFieldType(String field) throws InterchangeException {
	FieldDescriptor fd = source.getFieldDescriptor(field);
	String type = fd.getType(); 
	int thisFieldType = IS_OTHER;
	try {
	    Class cl = Class.forName(type);
	    if (Number.class.isAssignableFrom(cl))
		thisFieldType = IS_NUMBER;
	    else if (String.class.isAssignableFrom(cl))
		thisFieldType = IS_STRING;
	    else if (Date.class.isAssignableFrom(cl))
		thisFieldType = IS_DATE;
	}
	catch(Exception e) { } // Ignore any problems...
	return thisFieldType;
    }

    private void doBruteSort(Object[] a, int[] map, int lo, int hi) {
        if ((hi-lo) == 1) {
            if (compare(a[map[hi]], a[map[lo]]) < 0) {
                int T = map[lo];
                map[lo] = map[hi];
                map[hi] = T;
	    }
	}
        if ((hi-lo) == 2) {
            int pmin = compare(a[map[lo]], a[map[lo+1]]) < 0 ? lo : lo+1;
            pmin = compare(a[map[pmin]], a[map[lo+2]]) < 0 ? pmin : lo+2;
            if (pmin != lo) {
                int T = map[lo];
                map[lo] = map[pmin];
                map[pmin] = T;
            }
            doBruteSort(a, map, lo+1, hi);
        }
        if ((hi-lo) == 3) {
            int pmin = compare(a[map[lo]], a[map[lo+1]]) < 0 ? lo : lo+1;
            pmin = compare(a[map[pmin]], a[map[lo+2]]) < 0 ? pmin : lo+2;
            pmin = compare(a[map[pmin]], a[map[lo+3]]) < 0 ? pmin : lo+3;
            if (pmin != lo) {
                int T = map[lo];
                map[lo] = map[pmin];
                map[pmin] = T;
            }
            int pmax = compare(a[map[hi]], a[map[hi-1]]) > 0 ? hi : hi-1;
            pmax = compare(a[map[pmax]], a[map[hi-2]]) > 0 ? pmax : hi-2;
            if (pmax != hi) {
                int T = map[hi];
                map[hi] = map[pmax];
                map[pmax] = T;
            }
            doBruteSort(a, map, lo+1, hi-1);
        }
    }

    private void doQuickSort(Object[] a, int[] map, int lo0, int hi0) {

        int lo = lo0;
        int hi = hi0;

        if ((hi-lo) <= 3) {
            doBruteSort(a, map, lo, hi);
            //doSelectionSort(a, map, lo, hi);
            return;
        }

        Object mid = a[map[(lo + hi) / 2]];

	int lastLo = lo;
	int lastHi = hi;
        while (lo <= hi) {
	    lastLo = lo;
	    lastHi = hi;
            while (compare(a[map[lo]], mid) < 0  && lo < hi0) {
                lo++;
            }
            while (compare(mid, a[map[hi]]) < 0 && hi > lo0 ) {
                hi--;
            }
            if (lo < hi) {
                int T = map[lo];
                map[lo] = map[hi];
                map[hi] = T;
                hi--;
                lo++;
            }
	    else if ((lo == lastLo) && (hi == lastHi))
		break;
        }
        if (hi < lo) {
            int T = hi;
            hi = lo;
            lo = T;
        }
	else if ((lo == lastLo) && (hi == lastHi)) {
	    if (! ((lo == hi) && (lo == (lo0 + hi0)/2))) {
	        doSelectionSort(a, map, lo0, hi0);
	    }
	}

        doQuickSort(a, map, lo0, lo);
        doQuickSort(a, map, lo == lo0 ? lo+1 : lo, hi0);
    }

    private void doSelectionSort(Object[] a, int[] map, int lo0, int hi0) {
	Object minVal, testVal;
	int minIdx = 0;
	for (int i = lo0; i <= hi0; i++) {
	    minIdx = i;
	    minVal = a[map[i]];
	    int j = i + 1;
	    for (j = i + 1; j <= hi0; j++) {
		testVal = a[map[j]];
		if (compare(testVal, minVal) < 0) {
		    minIdx = j;
		    minVal = testVal;
		}
	    } 
	    if (minIdx != i ) {
	        int tmp = map[i];
		map[i] = map[minIdx];
		map[minIdx] = tmp;
	    }
	}
    }
  
    private synchronized void startSort() {
	if ((source != null) && (field != null)) {
	    Object[] data = null;
	    map = new int[0];

	    try {
		String[] flds = {field};
		data = source.get(flds, 0)[0];
		map = new int[source.getRecordCount()];
		for (int r = 0; r < map.length; r++)
		    map[r] = r;
System.out.println("Starting at: " + new java.util.Date());
		numCompares = 0;
		doQuickSort(data, map, 0, map.length - 1);
		//doSelectionSort(data, map, 0, map.length - 1);
System.out.println("Done at: " + new java.util.Date() + " Compares = " + numCompares);
	    }
	    catch (InterchangeException e) {
		e.printStackTrace();
	    }
	}
    }


    //     ---------------- Sampler methods ---------------- 
   /**
     * Translates the requested record number into the record number in
     * the underlying view.
     */
    public int getRecordNumber(int requestedRecord) throws
        NoSuchRecordException {

	if (source == null)
	    throw new NoSuchRecordException(requestedRecord);

	else if (field == null) 
	    return requestedRecord;

	else if ((requestedRecord < 0) || (requestedRecord >= map.length))
	    throw new NoSuchRecordException(requestedRecord);

	return map[requestedRecord];
    }
 
    /**
     * Returns the size of the sample. This is trivial for the sorted view,
     * since we aren't actually changing anything.
     */
    public synchronized int size() {
	if ((map == null) || (source == null))
	    return 0;
	else if (field == null) {
	    try {
		return source.getRecordCount();
	    }
	    catch (InterchangeException e) {
		return 0;
	    }
	}
	else
	    return map.length;
    }
 
    /**
     * Returns true if the sample includes the record number in the
     * underlying source view. This is used to filter out events which
     * indicate changes to records which are hidden by this sample.
     */
    public synchronized boolean recordIsInSample(int record) {
	return ((record >= 0) && (record < size()));
    }


}

