package sieve.sluice.views;

import sieve.sluice.*;
import sieve.sluice.event.*;
import java.util.*;
import java.awt.*;

/**
 * A TableSelectionView looks for a field of TableView objects in its
 * source view, allows selection on of these tables, and then behaves
 * like the selected table.
 * If the source view does not contain any fields of TableViews, then
 * the TableSelectionView behaves like its source.
 * Since switching the selected view may result in radically different
 * behavior, we have to take some precautions:
 * <UL>
 * <LI> Event listeners are registered with the TableSelectionView, and
 * the TableSelectionView listens to the currently selected view.
 * <LI> When the selected view is changed, the TableSelectionView
 * de-registers itself from the old selection, registers itself with
 * the new one, and fires an event to its (the TableSelectionView's) 
 * listeners.
 * </UL>
 */
public class TableSelectionView extends DefaultTableView implements TableView {

    /**
     * This contains what would normally be considered our source. 
     * So that we can re-use some of DefaultTableView's code, we'll
     * let our parent's source field point to the currently selected
     * view.
     */
    TableView   	tableOfTables   	= null; 

    FieldDescriptor	selectedField 		= null;

    TableSelectionViewListenerAdapter 	listener	= null;

    int			selectedTableIndex	= -1;

    /**
     * If this is true, then when a source view is added  and that
     * view does, in fact, contain a field of tables, then we  
     * automatically initialize ourselves by choosing the first table.
     * Otherwise we will throw NoSourceExceptions until a table
     * is actually selected.
     */
    boolean     initWithFirstTable = false;

    public TableSelectionView() {
	this(false);
    }

    /**
     * If initWithFirstTable is true, then when a source view is added 
     * and that view does, in fact, contain a field of tables, then we  
     * automatically initialize ourselves by choosing the first table.
     * Otherwise we will throw NoSourceExceptions until a table
     * is actually selected.
     */
    public TableSelectionView(boolean initWithFirstTable) {
	super();
	this.initWithFirstTable = initWithFirstTable;
    }

    public TableView getSelectedTableView() throws NoSourceException {
	if (tableOfTables == null)
	    throw new NoSourceException();
	return source;
    }

    public int getSelectedTableViewIndex() throws NoSourceException {
	if (tableOfTables == null)
	    throw new NoSourceException();

	return selectedTableIndex;
    }

    public void setSelectedTableViewIndex(int idx) throws InterchangeException {
	int oldIdx = idx;

	if (tableOfTables == null)
	    throw new NoSourceException();

	TableView tv = null;

	String[] fld = {selectedField.getName()};
	tv = (TableView) tableOfTables.get(fld, idx, 1)[0][0];

	this.source = tv;
	this.selectedTableIndex = idx;

	if (listener != null)
	    listener.stopListening();

	if (source != tableOfTables) 
            listener = new TableSelectionViewListenerAdapter(this, source);

	eventHandler.fireTableViewEvent(new TableViewEvent(this));
	changes.firePropertyChange("selectedTableViewIndex", 
	    null, new Integer(idx));
    }

    /**
     * Returns an array of views which are available for selection.
     */
    public TableView[] getAvailableTableViews() throws NoSourceException {
	if ((tableOfTables == null) || (selectedField == null))
	    throw new NoSourceException();

	TableView[] views = null;
	try {
	    String[] flds = {selectedField.getName()};
	    if (tableOfTables.getRecordCount() == 0)
		views = new TableView[0];
	    else {
	        Object[][] data = tableOfTables.get(flds, 0);
	        views = new TableView[data[0].length];
	        for (int i = 0; i < views.length; i++)
	            views[i] = (TableView)(data[0][i]);
	    }
	}
	catch (Exception e) {
	    e.printStackTrace();
	}
	return views;
    }


    /* -------------  Methods for link support ------------- */

    private void setSource(TableView tableOfTables) {
	this.tableOfTables = tableOfTables;
	FieldDescriptor oldTableField = selectedField;

	// For now, make the fields declare themselves as TableViews or Tables
	String[] types = {"sieve.sluice.TableView", "sieve.sluice.Table"};
	TypeView tv = new TypeView(tableOfTables, types);
	try {
	    String [] flds = tv.getFieldNames();
	    if (flds.length > 0) {
		// If we already had a field of tables, see if that field
		// still exists.
		if (oldTableField != null) {
		    try {
			selectedField = tableOfTables.getFieldDescriptor(
				oldTableField.getName());
			return;
		    }
		    catch(InterchangeException ex) {}// Ignore if not there
		}

		// Just use the first field of tables
		selectedField = tableOfTables.getFieldDescriptor(flds[0]);

		TableView[] views = getAvailableTableViews();
		if (views.length > 0) {
		    if (this.source != null) {
		        // Make sure the thing we think is our source is
		        // still in the table.
		        boolean found = false;
		        for (int i = 0; i < views.length && ! found; i++)	
			    found = (views[i] == this.source);
		        if (!found)
			    this.source = null;
		    }
		
		    if ((initWithFirstTable) && (this.source == null)) 
		        setSelectedTableViewIndex(0);
		    return;
		}
	    }
	}
	catch (InterchangeException e) {
	    e.printStackTrace();
	}
	// If anything goes wrong, then we'll just act like our source.
	this.source = tableOfTables;
    }

    public void addSourceView(TableView source) throws TooManySourcesException {
	if (tableOfTables != null)
	    throw new TooManySourcesException();
	setSource(source);
	eventHandler.fireTableViewEvent(new SourceAddedEvent(this));
    }

    public void removeSourceView(TableView source) {
	if (source == tableOfTables) {
	    if (listener != null) {
	        listener.stopListening();
	        listener = null;
	    }
	    eventHandler.fireTableViewEvent(new SourceRemovedEvent(this));
	}
    }


    /**
     * This method will be called by a TableSelectionViewListenerAdapter
     * when the selected view changes.
     */
    void selectedTableViewChanged(TableViewEvent event) {
        eventHandler.fireTableViewEvent(event);
    }

    /**
     * Need to think about what to do here. We probably need a different
     * kind of listener registration for events that affect the structure
     * of the table of tables.
     */
    public void tableViewChanged(TableViewEvent event) {
	// Check if we are passing our source table rather than a selected
	// table.
/*
	if (this.source == tableOfTables)
            eventHandler.fireTableViewEvent(event);
	else
	    setSource(tableOfTables);
*/
            eventHandler.fireTableViewEvent(event);
    }


}

/**
 * This adapter class is used to let us distinguish between events 
 * generated by our source (the table of tables) and those generated
 * by the currently selected view. 
 */
class TableSelectionViewListenerAdapter implements TableViewListener {
    TableSelectionView 	v 		= null;
    TableView 		selectedView 	= null;

    public TableSelectionViewListenerAdapter(TableSelectionView v, 
	TableView selectedView) {
	this.v = v;
	this.selectedView = selectedView;
	selectedView.addTableViewListener(this);
    }

    public void tableViewChanged(TableViewEvent event) {
	v.selectedTableViewChanged(event);
    }

    public void stopListening() {
	selectedView.removeTableViewListener(this);
    }

}

