package sieve.whiteboard;

import java.awt.*;
import java.awt.event.*;
import java.beans.*;
import sieve.workbench.*;
 
/**
 * A Doodle component. These are created by the DoodleTool.
 * @see DoodleTool
 */
public class Doodle extends Component implements WorkbenchComponent,
	MouseMotionListener, MouseListener {

    static {
	ToolFactory.getToolFactory().registerTool(Doodle.class, 
		DoodleTool.class);
    }

    public static final int HANDLE_SIZE = 3;

    private static int STATE_UNSELECTED = 0;
    private static int STATE_SELECTED   = 1;
    private static int STATE_MOVING     = 2;
    int state = STATE_UNSELECTED;

    Point start = null;
    Point end = null;
    Point lastPoint = null;
    Polygon poly = null;
    Workbench wb = null;

    protected Rectangle startHandle = new Rectangle();
    protected Rectangle startHandle2 = new Rectangle();
    protected Rectangle endHandle = new Rectangle();
    protected Rectangle endHandle2 = new Rectangle();

    UniqueID id = null;
    private PropertyChangeSupport listeners = null;

    private boolean constructed = false;

    private Color color = Color.black;
    private boolean filled = false;

    public synchronized UniqueID getUniqueID() {
	return id;
    }

    public synchronized void setUniqueID(UniqueID id) {
	this.id = id;
    }

    void addPoint(Point p) {
	if (!constructed) {
	    end = p;
	    poly.addPoint(p.x, p.y);
	}
    }

    public Doodle(Point start, Workbench wb) {
	poly = new Polygon();
	this.start = start;
	this.end = start;
	poly.addPoint(start.x, start.y);
	this.wb = wb;
	this.id = new UniqueID(this);
	addMouseListener(this);
	addMouseMotionListener(this);
	this.listeners = new PropertyChangeSupport(this);
    }

    public Rectangle getBounds() {
	Dimension parentSize = getParent().getSize();
        return new Rectangle(0, 0, parentSize.width, parentSize.height);
    }

    public void paint(Graphics g) {
        g.setColor(color);
	if (isSelected()) {
	    Rectangle bounds = poly.getBounds();
	    startHandle.setBounds(bounds.x - HANDLE_SIZE,
                bounds.y - HANDLE_SIZE,
                2 * HANDLE_SIZE, 2 * HANDLE_SIZE);
            startHandle2.setBounds(bounds.x + bounds.width - HANDLE_SIZE,
                bounds.y - HANDLE_SIZE,
                2 * HANDLE_SIZE, 2 * HANDLE_SIZE);
            endHandle.setBounds(bounds.x + bounds.width - HANDLE_SIZE, 
		bounds.y + bounds.height - HANDLE_SIZE,
                2 * HANDLE_SIZE, 2 * HANDLE_SIZE);
            endHandle2.setBounds(bounds.x - HANDLE_SIZE,
                bounds.y + bounds.height - HANDLE_SIZE,
                2 * HANDLE_SIZE, 2 * HANDLE_SIZE);
            drawHandle(g, startHandle);
            drawHandle(g, startHandle2);
            drawHandle(g, endHandle);
            drawHandle(g, endHandle2);
	}

        if ((start != null) && (end != null)) {
	    if (filled)
                g.fillPolygon(poly);
	    else
                g.drawPolyline(poly.xpoints, poly.ypoints, poly.npoints);
	}
    }

    private void drawHandle(Graphics g, Rectangle handle) {
        g.fillRect(handle.x, handle.y, handle.width, handle.height);
    }

    public Color getColor() {
	return color;
    }
    public void setColor(Color color) {
	this.color = color;
	repaint();
    } 
    public boolean getFilled() {
	return filled;
    }
    public void setFilled(boolean filled) {
	this.filled = filled;
	repaint();
    } 

    public boolean contains(int x, int y) {
	if (!constructed)
	    return false;
	return poly.contains(x, y) ||
	    (isSelected() && 
	    (startHandle.contains(x, y) ||
	    startHandle2.contains(x, y) ||
	    endHandle.contains(x, y) ||
	    endHandle2.contains(x, y)));
    }

    public boolean contains(Point p) {
	return contains(p.x, p.y);
    }

    void setConstructed(boolean constructed) {
	this.constructed = constructed;
    }

    // Use for drawing handles only?
    public void setColorSource(ColorSource cs) {
    }

    public Object getModel() {
	return this;
    }

    public Component getView() {
	return this;
    }

    public void remove() {
        wb.remove(this);
        wb.componentRemoved(this);
    }

    public boolean isSelected() {
	return (state != STATE_UNSELECTED);
    }

    public void setSelected(boolean selected) {
	if (isSelected() != selected) {
	    if ((selected) && (wb.requestSelection(this))) {
		wb.setCurrentSelection(this);
		this.state = STATE_SELECTED;
		wb.requestEditor(this);
	    }
	    else {
		this.state = STATE_UNSELECTED;
		if (wb.getCurrentSelection() == this) 
		    wb.setCurrentSelection(null);
		lastPoint = null;
	    }
	    repaint();
     	}
    }

    public void addPropertyChangeListener(PropertyChangeListener l) {
	listeners.addPropertyChangeListener(l);
    }
    public void removePropertyChangeListener(PropertyChangeListener l) {
	listeners.removePropertyChangeListener(l);
    }

    public synchronized Point getStart() {
	return start;
    }
    public synchronized void setStart(Point start) {
	Point orig = new Point(this.start);
	if (! orig.equals(start)) {
	    int xDiff = start.x - this.start.x;
	    int yDiff = start.y - this.start.y;
	    poly.translate(xDiff, yDiff);
	    this.start = start;
 	    repaint();
	    listeners.firePropertyChange("start", orig, new Point(start));
	}
    }

    private boolean onHandle(Point endPoint, Point p) {
	return (Math.abs(endPoint.x - p.x) < 6) && 
		(Math.abs(endPoint.y - p.y) < 6);
    }

    public void mouseMoved(MouseEvent e) { }

// This needs to fire propChangeEvents. Note that WB.componentBoundsChanged
// is NOT appropriate, since the bounds of a line are always the same.
    public synchronized void mouseDragged(MouseEvent e) {
	if (isSelected()) {
	    if (lastPoint != null) {
		Point origStart = new Point(start);
		Point origEnd = new Point(end);
		int xDiff = e.getPoint().x - lastPoint.x;
		int yDiff = e.getPoint().y - lastPoint.y;
		if (state == STATE_MOVING) {
		    poly.translate(xDiff, yDiff);
		    start.x += xDiff;
		    start.y += yDiff;
	    	    listeners.firePropertyChange("start", origStart, 
			new Point(start));
		    end.x += xDiff;
		    end.y += yDiff;
		}
	    }
	    lastPoint = e.getPoint();
	    repaint();
	}
    }

    public void mousePressed(MouseEvent e) {
   	setSelected(true);
	lastPoint = e.getPoint();
	if (isSelected()) 
	    state = STATE_MOVING;
    }
    public void mouseReleased(MouseEvent e) {
	if (isSelected())
	    state = STATE_SELECTED;
	lastPoint = null;
    }
    public void mouseClicked(MouseEvent e) {}
    public void mouseEntered(MouseEvent e) {}
    public void mouseExited(MouseEvent e) {}

}


