package sieve.whiteboard;

import java.awt.*;
import java.awt.event.*;
import java.beans.*;
import sieve.workbench.*;
 
/**
 * An abstract RectangularComponent for things like boxes, ovals, and
 * doodles.
 */
public class RectangularComponent extends Component implements 
	WorkbenchComponent, MouseMotionListener, MouseListener {

    static {
	WorkbenchScaler.addScaler(RectangularComponent.class, 
	    new WhiteBoardScaler());
    }

    public static final int HANDLE_SIZE = 3;

    private static int STATE_UNSELECTED = 0;
    private static int STATE_SELECTED   = 1;
    private static int STATE_MOVING     = 2;
    private static int STATE_RESHAPING  = 3;
    int state = STATE_UNSELECTED;

    Rectangle handleBeingMoved = null;

    Point start = null;
    Point end = null;
    Point lastPoint = null;
    Polygon poly = null;
    Workbench wb = null;

    UniqueID id = null;
    private PropertyChangeSupport listeners = null;

    protected boolean constructed = false;
    protected Rectangle startHandle = new Rectangle();
    protected Rectangle startHandle2 = new Rectangle();
    protected Rectangle endHandle = new Rectangle();
    protected Rectangle endHandle2 = new Rectangle();

    private Color color = Color.black;

    public synchronized UniqueID getUniqueID() {
	return id;
    }

    public synchronized void setUniqueID(UniqueID id) {
	this.id = id;
    }

    public RectangularComponent(Point start, Point end, Workbench wb) {
	this.start = start;
	this.end = end;
	this.wb = wb;
	this.id = new UniqueID(this);
	addMouseListener(this);
	addMouseMotionListener(this);
	this.listeners = new PropertyChangeSupport(this);
    }

    public Rectangle getBounds() {
	Dimension parentSize = getParent().getSize();
        return new Rectangle(0, 0, parentSize.width, parentSize.height);
    }

    public void setColor(Color color) {
	if ((this.color != null) && ! color.equals(this.color)) {
	    this.color = color;
	    listeners.firePropertyChange("color", null, color);
	}
    }
    public Color getColor() {
	return color;
    }

    public void paint(Graphics g) {
	if (isSelected()) {
	    startHandle.setBounds(start.x - HANDLE_SIZE, 
		start.y - HANDLE_SIZE, 
		2 * HANDLE_SIZE, 2 * HANDLE_SIZE);
	    startHandle2.setBounds(end.x - HANDLE_SIZE, 
		start.y - HANDLE_SIZE, 
		2 * HANDLE_SIZE, 2 * HANDLE_SIZE);
	    endHandle.setBounds(end.x - HANDLE_SIZE, end.y - HANDLE_SIZE, 
		2 * HANDLE_SIZE, 2 * HANDLE_SIZE);
	    endHandle2.setBounds(start.x - HANDLE_SIZE, 
		end.y - HANDLE_SIZE, 
		2 * HANDLE_SIZE, 2 * HANDLE_SIZE);
	    drawHandle(g, startHandle);
	    drawHandle(g, startHandle2);
	    drawHandle(g, endHandle);
	    drawHandle(g, endHandle2);
	}
    }

    private void drawHandle(Graphics g, Rectangle handle) {
	g.fillRect(handle.x, handle.y, handle.width, handle.height);
    }

    public boolean contains(int x, int y) {
	if ((!constructed) || (isSelected() && 
	    (startHandle.contains(x, y) || endHandle.contains(x, y) ||
	    startHandle2.contains(x, y) || endHandle2.contains(x, y))))
	    return true;

	Rectangle bounds = new Rectangle(Math.min(start.x, end.x), 
	     	Math.min(start.y, end.y), Math.abs(end.x - start.x) + 1, 
		Math.abs(end.y - start.y) + 1);
	Rectangle innerBounds = new Rectangle(Math.min(start.x, end.x) + 1, 
	     	Math.min(start.y, end.y) + 1, Math.abs(end.x - start.x) - 2, 
		Math.abs(end.y - start.y) - 2);
	return bounds.contains(x, y) && !innerBounds.contains(x, y);
    }

    public boolean contains(Point p) {
	return contains(p.x, p.y);
    }

    void setConstructed(boolean constructed) {
	this.constructed = constructed;
    }

    // Use for drawing handles only?
    public void setColorSource(ColorSource cs) {
    }

    public Object getModel() {
	return this;
    }

    public Component getView() {
	return this;
    }

    public void remove() {
	wb.remove(this);
	wb.componentRemoved(this);
    }

    public boolean isSelected() {
	return (state != STATE_UNSELECTED);
    }

    public void setSelected(boolean selected) {
	if (isSelected() != selected) {
	    if ((selected) && (wb.requestSelection(this))) {
		wb.setCurrentSelection(this);
		this.state = STATE_SELECTED;
		wb.requestEditor(this);
	    }
	    else {
		this.state = STATE_UNSELECTED;
		this.handleBeingMoved = null;
		if (wb.getCurrentSelection() == this) 
		    wb.setCurrentSelection(null);
		lastPoint = null;
	    }
	    repaint();
     	}
    }

    public void addPropertyChangeListener(PropertyChangeListener l) {
	listeners.addPropertyChangeListener(l);
    }
    public void removePropertyChangeListener(PropertyChangeListener l) {
	listeners.removePropertyChangeListener(l);
    }

    public synchronized Point getStart() {
	return start;
    }
    public synchronized void setStart(Point start) {
	Point orig = new Point(this.start);
	if (! orig.equals(start)) {
	    this.start = start;
 	    repaint();
	    listeners.firePropertyChange("start", orig, new Point(start));
	}
    }
    public synchronized Point getEnd() {
	return end;
    }
    public synchronized void setEnd(Point end) {
	Point orig = new Point(this.end);
	if (! orig.equals(end)) {
	    this.end = end;
 	    repaint();
	    listeners.firePropertyChange("end", orig, new Point(end));
	}
    }

    private boolean onHandle(Point endPoint, Point p) {
	return (Math.abs(endPoint.x - p.x) < 6) && 
		(Math.abs(endPoint.y - p.y) < 6);
    }

    public void mouseMoved(MouseEvent e) { }

    public synchronized void mouseDragged(MouseEvent e) {
	if (isSelected()) {
	    if (lastPoint != null) {
		Point origStart = new Point(start);
		Point origEnd = new Point(end);
		int xDiff = e.getPoint().x - lastPoint.x;
		int yDiff = e.getPoint().y - lastPoint.y;
		if (state == STATE_RESHAPING) {
		    if (handleBeingMoved == startHandle) {
		        start.x += xDiff;
		        start.y += yDiff;
	    	        listeners.firePropertyChange("start", origStart, 
			    new Point(start));
		    }
		    else if (handleBeingMoved == endHandle) {
		        end.x += xDiff;
		        end.y += yDiff;
	    	        listeners.firePropertyChange("end", origEnd, 
			    new Point(end));
		    }
		    else if (handleBeingMoved == endHandle2) {
		        start.x += xDiff;
		        end.y += yDiff;
	    	        listeners.firePropertyChange("end", origEnd, 
			    new Point(end));
	    	        listeners.firePropertyChange("start", origStart, 
			    new Point(start));
		    }
		    else if (handleBeingMoved == startHandle2) {
		        end.x += xDiff;
		        start.y += yDiff;
	    	        listeners.firePropertyChange("end", origEnd, 
			    new Point(end));
	    	        listeners.firePropertyChange("start", origStart, 
			    new Point(start));
		    }
		}
		else if (state == STATE_MOVING) {
		    start.x += xDiff;
		    start.y += yDiff;
	    	    listeners.firePropertyChange("start", origStart, 
		        new Point(start));
		    end.x += xDiff;
		    end.y += yDiff;
	    	    listeners.firePropertyChange("end", origEnd, 
			new Point(end));
		}
	    }
	    lastPoint = e.getPoint();
	    repaint();
	}
    }

    public void mousePressed(MouseEvent e) {
   	setSelected(true);
	lastPoint = e.getPoint();
	if (isSelected()) {
	    if (endHandle.contains(lastPoint) && 
		!startHandle.contains(lastPoint)) {
		state = STATE_RESHAPING;
		handleBeingMoved = endHandle;
	    }
	    else if (startHandle.contains(lastPoint) && 
		!endHandle.contains(lastPoint)) {
		state = STATE_RESHAPING;
		handleBeingMoved = startHandle;
	    }
	    else if (startHandle2.contains(lastPoint)) {
		state = STATE_RESHAPING;
		handleBeingMoved = startHandle2;
	    }
	    else if (endHandle2.contains(lastPoint)) {
		state = STATE_RESHAPING;
		handleBeingMoved = endHandle2;
	    }
	    else
		state = STATE_MOVING;
	}
    }
    public void mouseReleased(MouseEvent e) {
	if (isSelected())
	    state = STATE_SELECTED;
	lastPoint = null;
	handleBeingMoved = null;
    }
    public void mouseClicked(MouseEvent e) {}
    public void mouseEntered(MouseEvent e) {}
    public void mouseExited(MouseEvent e) {}

}


