package sieve.workbench;

import java.awt.*;
import java.awt.event.*;
import java.beans.*;
import java.lang.reflect.*;

public class DefaultWrapper extends Container implements WorkbenchComponent,
    MouseListener, MouseMotionListener, PropertyChangeListener {
    // Shorthands for the cursors.
    private static Cursor nwResizeCursor = 
	Cursor.getPredefinedCursor(Cursor.NW_RESIZE_CURSOR);
    private static Cursor neResizeCursor = 
	Cursor.getPredefinedCursor(Cursor.NE_RESIZE_CURSOR);
    private static Cursor swResizeCursor = 
	Cursor.getPredefinedCursor(Cursor.SW_RESIZE_CURSOR);
    private static Cursor seResizeCursor = 
	Cursor.getPredefinedCursor(Cursor.SE_RESIZE_CURSOR);
    private static Cursor moveCursor = 
	Cursor.getPredefinedCursor(Cursor.MOVE_CURSOR);
    private static Cursor defaultCursor = 
	Cursor.getPredefinedCursor(Cursor.CROSSHAIR_CURSOR);
	// Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR);

    ColorSource cs = null;
    Workbench wb = null;
    boolean selected = false;
    Object model = null;
    Component view = null;
    int borderWidth = 7;
    int resizeDelta = 10; // The size of the corners (for resizing)
   
    // The state of the cursor outside of this wrapper
    Cursor resetCursor = null; 

    // The last point which was pressed or dragged
    Point lastPoint = null;

    // ID of this component
    UniqueID id = null;

    // Whether or not to show border around unselected objects
    private static boolean showBorder = true;

    public DefaultWrapper(Workbench wb, Object model) {
	this.wb = wb;
	this.model = model;
	if (model instanceof Component)
	    view = (Component) model;
	else
	    view = getCustomizer(model);
	setLayout(null);
	add(view);
	addMouseListener(this);
	addMouseMotionListener(this);

 	this.id = new UniqueID(this);

	// See if the object has bound properties...
	try {
                Class[] args = { PropertyChangeListener.class };
                Method registerMethod =
                    model.getClass().getMethod("addPropertyChangeListener",
                    args);
                Object[] me = { this };
                registerMethod.invoke(model, me);
        }
        catch(Exception e){} // Fail quietly -- there may be no bound props
	setCursor(defaultCursor);
    }

    public UniqueID getUniqueID() {
	return this.id;
    }

    public void setUniqueID(UniqueID id) {
	this.id = id;
    }

    public void setColorSource(ColorSource cs) {
	this.cs = cs;
	repaint();
    }

    public Object getModel() {
	return model;
    }

    public Component getView() {
	//return view;
	return this;
    }

    public void remove() {
	wb.remove(this);
	wb.componentRemoved(this);
    }

    public void setSelected(boolean selected) {
	if (selected != this.selected) {
	    boolean isWBsSelection = (wb.getCurrentSelection() == this);
	    if ((selected) && !isWBsSelection && (wb.requestSelection(this))) {

		setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
                this.selected = true;

		if (!isWBsSelection)
                    wb.setCurrentSelection(this);

	    	repaint();

		// Since building the editor is slow, we'll do it in a
		// separate thread. This should possibly move to either
		// the Workbench class, or to the DefaultPropertyEditor
		// class. (Probably the latter.)
		final WorkbenchComponent wbc = this;
		Thread t = new Thread() {
		    public void run() {
		        wb.requestEditor(wbc);
		    }
		};
		t.setPriority(Thread.MIN_PRIORITY);
		t.start();

		setCursor(defaultCursor);
            }
	    else {
		this.selected = false;
		if (wb.getCurrentSelection() == this)
                    wb.setCurrentSelection(null);
	    	repaint();
	    }
	}
    }

    public boolean isSelected() {
	return selected;
    }

    public static void setShowBorders(boolean showBorder) {
	DefaultWrapper.showBorder = showBorder;
    }

    public static boolean getShowBorders() {
	return showBorder;
    }

    public void paint(Graphics g) {
	if (cs != null)
	    g.setColor(cs.getColor());
	else
	    g.setColor(Color.gray);

	if (selected) 
	    g.fillRect(0, 0, getSize().width - 1, getSize().height - 1);
	else if (showBorder) {
	    g.drawRect(0, 0, getSize().width - 1, getSize().height - 1);
	    g.drawRect(borderWidth - 1, borderWidth - 1, 
	        getSize().width - 2 * borderWidth + 1,
	        getSize().height - 2 * borderWidth + 1);
	}
	g.setColor(wb.getBackground());
	g.fillRect(borderWidth, borderWidth, 
	    getSize().width - 1 - 2 * borderWidth,
	    getSize().height - 1 - 2 * borderWidth);
	super.paint(g);
    }

    public Dimension getPreferredSize() {
	Dimension viewSize = view.getPreferredSize();
	if (viewSize.width <= 0 || viewSize.height <= 0)
	    viewSize = view.getMinimumSize();
	int width = Math.max(viewSize.width, 2) + (2 * borderWidth);
        int height = Math.max(viewSize.height, 2) + (2 * borderWidth);
        return new Dimension(width, height);
    }

    public void addNotify() {
	super.addNotify();
	view.setLocation(new Point(borderWidth, borderWidth));
	view.setSize(view.getPreferredSize());
	setSize(getPreferredSize());
	if (model instanceof java.applet.Applet) {
	    final java.applet.Applet app = (java.applet.Applet) model;
	    Thread t = new Thread() {
		public void run() {
	    	    app.start();
		}
	    };
	    t.start();
	}
	sizeToFit();
    }

    public void layout() {
	sizeToFit();
    }

    private void sizeToFit() {
	setSize(getPreferredSize());
	view.repaint();
    }

    public void setBounds(int x, int y, int width, int height) {
	super.setBounds(x, y, width, height);
        view.setBounds(borderWidth, borderWidth,
                width-(2*borderWidth), height-(2*borderWidth));
        view.validate();
    }

    private Component getCustomizer(Object bean) {
        Container cust = new Panel();
        cust.setLayout(new BorderLayout());
        Component customizer = null;
        try {
            BeanInfo bi = Introspector.getBeanInfo(bean.getClass());
            BeanDescriptor bd = bi.getBeanDescriptor();
            Class customizerClass = bd.getCustomizerClass();
            Label header = new Label(bd.getDisplayName());
            header.setAlignment(Label.CENTER);
            cust.add("North", header);
 
            if (customizerClass == null) {
                // Property panel will listen for changes, so we
                // don't have to
                customizer = new PropertyPanel(bean, this, wb);
            }
            else {
                customizer = (Component)customizerClass.newInstance();
                ((Customizer)customizer).setObject(bean);
                //((Customizer)customizer).addPropertyChangeListener(this);
            }
            cust.add("Center", customizer);
            Dimension size = customizer.getPreferredSize();
            if (cust.getParent() != null)
                cust.getParent().layout();
        }
        catch (Exception e) {
            e.printStackTrace();
        }
        return cust;
    }

    
    // ------------------ PropertyChangeListener methods ------------------

    public void propertyChange(PropertyChangeEvent e) {
	wb.componentEdited(this, e);
    }


    // ---------------------- MouseListener methods -----------------------
    private Cursor getAppropriateCursor(MouseEvent e) {
        int x = e.getX();
        int y = e.getY();
        int lowX = borderWidth;
        int highX = getSize().width - borderWidth;
        int lowY = borderWidth;
        int highY = getSize().height - borderWidth;
	Cursor c = null;

	// Check if we're on the border.
	if (x < lowX || x > highX || y < lowY || y > highY) {
            // We're on the border.  If we're in a corner set the cursor
            // to indicate a resize, otherwise a move.
            lowX += resizeDelta;
            lowY += resizeDelta;
            highX -= resizeDelta;
            highY -= resizeDelta;

            if (x < lowX && y < lowY) {
                c = (nwResizeCursor);
            } else if (x < lowX && y > highY) {
                c = (swResizeCursor);
            } else if (x > highX && y < lowY) {
                c = (neResizeCursor);
            } else if (x > highX && y > highY) {
                c = (seResizeCursor);
            } else {
                c = (moveCursor);
            }
	}
	else
	    c = (defaultCursor);
	return c;
    }

    public void mouseClicked(MouseEvent e) { }
    public void mousePressed(MouseEvent e) { 
	if (isSelected()) 
	    lastPoint = e.getPoint();
    }
    public void mouseReleased(MouseEvent e) { 
	Cursor c = getAppropriateCursor(e);
	if (c != defaultCursor) {
	    if (e.getModifiers() == Event.META_MASK) {
		try {
		    Link link = new LinkProxy(wb, this);
		    if (wb.requestLinkStart(this, link))  {
		        link.getView().setBounds(0, 0, wb.getSize().width, 
			    wb.getSize().height);
		        wb.add(link.getView());
		    }
		    else
	                setSelected(true);
		}
		catch(LinkException le) {
		    System.out.println("Couldn't create link: " + 
			le.getMessage());
	            setSelected(true);
		}
	    }
	    else 
	        setSelected(true);
	}
	lastPoint = null;
    }
    public void mouseEntered(MouseEvent e) { 
	if (isSelected() && (lastPoint == null)) {
	    resetCursor = getCursor();
	    setCursor(getAppropriateCursor(e));
	}
    }
    public void mouseExited(MouseEvent e) { 
	if ((resetCursor != null) && (lastPoint == null)) {
	    setCursor(resetCursor);
	    resetCursor = null;
	}
    }
    public void mouseMoved(MouseEvent e) { 
	if (isSelected()) {
	    setCursor(getAppropriateCursor(e));
	}
    }
    public void mouseDragged(MouseEvent e) { 
	if (isSelected()) {
	    if (lastPoint != null) {
	        Cursor c = getCursor();
		int xDiff = e.getPoint().x - lastPoint.x;
                int yDiff = e.getPoint().y - lastPoint.y;
		if (c == moveCursor) {
		    Point p = getLocation();
		    p.x += xDiff;
		    p.y += yDiff;
		    setLocation(p);
		    wb.componentBoundsChanged(this);
		}
		else if (c != defaultCursor) {
		    Rectangle r = getBounds();
		    if (c == nwResizeCursor) {
		        r.x += xDiff;
		    	r.width -= xDiff;
		    	r.y += yDiff;
		    	r.height -= yDiff;
		    }
		    else if (c == neResizeCursor) {
		    	r.width += xDiff;
		        lastPoint.x += xDiff;
		    	r.y += yDiff;
		    	r.height -= yDiff;
		    }
		    else if (c == seResizeCursor) {
		        lastPoint.x += xDiff;
		        lastPoint.y += yDiff;
		    	r.width += xDiff;
		    	r.height += yDiff;
		    }
		    else if (c == swResizeCursor) {
		        lastPoint.y += yDiff;
			r.x += xDiff;
		    	r.width -= xDiff;
		    	r.height += yDiff;
		    }
		    setBounds(r);
		    wb.componentBoundsChanged(this);
		}
		else {
	    	    lastPoint = e.getPoint();
		}

	    }
	    else
	        lastPoint = e.getPoint();
	}
    }
    // --------------------------------------------------------------------
}

