package sieve.workbench;

import java.awt.event.MouseMotionListener;
import java.io.Serializable;
import java.awt.Component;

public interface Link extends MouseMotionListener {

    /**
     * Initializes a Link with the given source, destination, and 
     * (optional) state data. (This is primarily intended for use by
     * the collaboration subsystem.) In order to be generically constructed
     * and initialized with this method, Link instances should implement
     * a parameter-less constructor.
     */
    public void init(Workbench wb, WorkbenchComponent source, 
	WorkbenchComponent dest, Serializable state) throws LinkException;

    /**
     * Returns the unique ID associated with this link.
     */
    public UniqueID getUniqueID();

    /**
     * Sets the unique ID associated with this link.
     */
    public void setUniqueID(UniqueID id);

    /**
     * Return the component from which the link was started.
     */
    public WorkbenchComponent getSourceComponent();

    /**
     * Return the component to which the link was drawn.
     */
    public WorkbenchComponent getDestComponent();

    /**
     * Called by the workbench when a component requests selection
     * when the link is being created. The Link should thow an exception if
     * it cannot use the component as its destination. 
     * The Workbench will assume that link creation has not
     * finished. Whether the dest is accepted or not, the component's 
     * request for selection will be denied.
     */
    public void setDestComponent(WorkbenchComponent c) throws LinkException; 

    /**
     * Returns a Serializable object representing the state of the Link.
     * This is optional if the Link has no state information besides
     * its source and dest. If it returns a non-null object, the only 
     * requirement is that the Link's <CODE>init</CODE> method should be
     * able to restore the Link based on the source, destination, and 
     * state object.
     */
    public Serializable getState();

    /**
     * Returns the component used to draw this link.
     */
    public Component getView();

    /**
     * Removes the link. The link should remove itself and, if it was
     * complete (i.e., not being constructed) should tell the Workbench
     * that it has gone away (by calling Workbench.linkRemoved).
     */
    public void remove();

}


