package sieve.workbench.collab;

import sieve.workbench.*;
import sieve.workbench.event.*;
import sieve.workbench.collab.msg.*;
import java.util.*;
import java.awt.Rectangle;
import java.awt.event.MouseEvent;
import java.io.Serializable;

/**
 * FullReplayRecorder is a ChannelRecorder that provides support for 
 * channel persistence and late joiners by 
 * storing (almost) all messages and returning a re-playable array of 
 * messages upon request. WorkbenchParticipant change messages (telepointer
 * changes) are ignored, as are all but the last bounds message for
 * any component.
 *
 * @see ChannelRecorder
 */
public class FullReplayRecorder implements ChannelRecorder, Serializable {
    private String name = null;

    private Vector messages = null;

    /**
     * We track the WorkbenchParticipant objects and ignore location
     * changes. This should also keep track of the last time a message
     * from each participant arrives, and prune the list if it
     * appears someone has died.
     */
    private Vector participants = null;

    /**
     * Track bounds messages and only send the last one
     */
    private Hashtable boundsMessages = null;

    public FullReplayRecorder(String name) {
	messages = new Vector();
	participants = new Vector();
	boundsMessages = new Hashtable();
	this.name = name;
    }

    public String getName() {
	return name;
    }

    public int getMessageCount() {
	return messages.size();
    }

    public synchronized Message[] getMessages() {
	Message[] msgs = new Message[messages.size()];
	messages.copyInto(msgs);
	return msgs;
    }

    public synchronized void doAddComponent(String className, 
	MouseEvent[] evts, UniqueID id) {
	if (className.equals(WorkbenchParticipant.class.getName()))
	    participants.addElement(id);
	Message m = new AddComponentMessage(id, className, evts);
	messages.addElement(m);
    }

    public synchronized void doRemoveComponent(UniqueID id) {
	Message m = new RemoveComponentMessage(id);
	messages.addElement(m);
    }

    public synchronized void doSetBounds(UniqueID id, Rectangle b) {
	Message oldBounds = (Message) boundsMessages.get(id);
	if (oldBounds != null)
	    messages.removeElement(oldBounds);
	Message m = new BoundsMessage(id, b);
	messages.addElement(m);
	boundsMessages.put(id, m);
    }
 
    public synchronized void doMakeLink(String linkClassName, UniqueID srcID, 
	UniqueID destID, UniqueID linkID, Serializable state) {
	Message m = new MakeLinkMessage(linkClassName, srcID, destID, linkID, 
	    state);
	messages.addElement(m);
    }
 
    public synchronized void doRemoveLink(UniqueID linkID) {
	Message m = new RemoveLinkMessage(linkID);
	messages.addElement(m);
    }
 
    public synchronized void doEditComponent(UniqueID objID, EditEvent e) {
	if (! ((e instanceof PropertyEditEvent) &&
	    (((PropertyEditEvent)e).getPropertyName().equals("location")) &&
	         participants.contains(objID) )) {
	    Message m = new EditMessage(objID, e);
	    messages.addElement(m);
	}
    }

    private class MessageRecord {
	private Message m = null;
	private UniqueID id = null;
	public MessageRecord(Message m, UniqueID id) {
	    this.m = m;
	    this.id = id;
	}
    }

}

