package sieve.workbench.collab;

import SharedData.*;
import java.net.*;
import java.util.*;
import sieve.workbench.collab.msg.*;

/**
 * This class (based on sieve.crew.RemoteControl) implements the 
 * WorkbenchClient, SharedData.ChannelConsumer and SharedData.Client 
 * interfaces. 
 */
public class JSDAClient 
implements ChannelConsumer, Client, WorkbenchClient, Runnable {

    protected Session session = null;
    protected Channel channel = null;
    protected String name = null;
    protected String channelName = null;
    protected SendQueue sendQueue = null;
    protected Thread life = null;

    private Integer lock = new Integer(0);
    private Integer qLock = new Integer(0);
    private WorkbenchController controller = null;
    private boolean okToReceive = false;
    private int messageCount = 0;
    private Vector rcvQueue = null;
    private boolean running = true;

    public JSDAClient(Session session, String channelName, 
	WorkbenchController controller) {

      	this.session = session;
    	this.channelName = channelName;
  
    	String localHostName = "LocalHost";
    	try {
      	    localHostName = InetAddress.getLocalHost().getHostName();
    	} 
 	catch (UnknownHostException uhex) {
	    uhex.printStackTrace();
        }
        name = "JSDAClient-"+localHostName+"-"+System.currentTimeMillis();
        joinSession();
        joinChannel();
	this.sendQueue = new SendQueue(session, channel, this, this);
	this.controller = controller;
	this.rcvQueue = new Vector();
	Thread life = new Thread(this);
	life.start();
    }

    private void joinSession() {
    	try {
      	    session.join(this);
      	    session.createChannel(this, channelName, true, true);
        } 
	catch (Exception sdex) {
            // this channel may already exist.  If so, great.
      	    if (sdex instanceof SharedDataException){
                if (((SharedDataException)sdex).getType()
            	    != SharedDataException.CHANNEL_EXISTS) {
                    sdex.printStackTrace();
            	}
      	    }
	    else
                sdex.printStackTrace();
        }
    }

    private void joinChannel() {
    	try {
            channel = Channel.join(session, channelName, this);
    	} 
	catch (SharedDataException sdex) {
      	    System.out.println("Error joining channel: "+sdex);
       	    sdex.printStackTrace();
        }
 
    	try {
      	    channel.addConsumer(this, this);
      	} 
	catch (SharedDataException e) {
      	    e.printStackTrace();
    	}
    }

    public void shutdown() {
	synchronized(lock) {
    	    try {
		running = false;
		synchronized(qLock) {
		    qLock.notifyAll();
		}
		sendQueue.waitForShutdown();
      	        channel.deleteConsumer(this, this);
      	        channel.leave(this);
	        channel = null; // Prevent further sends
      	        session.leave(this);
    	    } 
	    catch (SharedDataException sdex) {
      	        System.out.println("Error shutting down client");
      	        sdex.printStackTrace();
    	    }
	}
    }

      // ----------- Client Interface methods ---------

    public String authenticate(AuthenticationInfo info) {
        return null;
    }
      
    public String getName() {
        return name;
    }

    public void clientChanged(ClientEvent event) { }

    // ------------------------------------------------ 


    public Channel getChannel() {
	return channel;
    }

    public void send(Message m) {
	if (sendQueue.size() > 25) {
	    if (m instanceof TransientPropertyChangeMessage) {
	        System.out.println("Warning: " + sendQueue.size() + 
			" unsent messages -- purging...");
		sendQueue.purgeTransientMessages();
	        System.out.println("    now: " + sendQueue.size() + 
			" unsent messages"); 
	    }
	}
	sendQueue.addElement(m);
    }


    public void setOKToDeliverMessages(boolean ok) {
	this.okToReceive = ok;
	synchronized(qLock) {
	    qLock.notifyAll();
	}
    }
    public boolean isOKToDeliverMessages() {
	return this.okToReceive;
    }

    public void run() {
	while (running) {
	    while (!okToReceive || (rcvQueue.size() == 0)) {
	        try {
	            synchronized(qLock) {
	                qLock.wait(10 * 1000);
		    }
		    if (!running)
			return;
	        }
	        catch(Exception e) {}
	    }
	    Message m = null;
	    synchronized(rcvQueue) {
	        m = (Message) rcvQueue.elementAt(0);
	        rcvQueue.removeElementAt(0);
	    }
	    m.deliver(controller);
	}
    }

    public void dataReceived(Data data) {
	Message m = Message.getMessage(data);
	rcvQueue.addElement(m);
	synchronized(qLock) {
	    qLock.notifyAll();
	}
    }

    private class SendQueue implements Runnable {
	private Vector queue = null;
     	private Integer sendQueueLock = new Integer(0);
     	private Integer shutdownLock = new Integer(0);
	private Thread life = null;
        private Session session = null;
	private Channel channel = null;
 	private Client client = null;
   	private Object sessionLock = null;
	private boolean running = true;
	private boolean waitingForShutdown = false;

        public SendQueue(Session session, Channel channel, Client client,
	    Object sessionLock) {
	    this.session = session;
	    this.channel = channel;
	    this.client = client;
	    this.sessionLock = sessionLock;
	    queue = new Vector();
	    life = new Thread(this);
	    life.setDaemon(true);
	    life.start();
        } 

        public int size() {
	    return queue.size();
        }

	public void waitForShutdown() {
	    waitingForShutdown = true;
	    while (queue.size() > 0)
		try {
		    synchronized(shutdownLock) {
		    	shutdownLock.wait();
		    }
		}
		catch (Exception e){}
	    running = false;
 	}

	// This could also check for EditEvents (in EditMessages) which 
	// are obsolete
        public void purgeTransientMessages() {
	    synchronized(queue) {
		int i = 0;
		while (i < queue.size() - 1) { // Never completely empty the Q
		    if (queue.elementAt(i) instanceof 
			TransientPropertyChangeMessage) 
			queue.removeElementAt(i);
		    else
			i++;
		}
	    }
        }

        public void addElement(Message m) {
	    if (running) {
	        queue.addElement(m);
	        synchronized(sendQueueLock) {
	            sendQueueLock.notifyAll();
	        }
	    }
      	}

        public void run() {
	    while (running) {

	        while (queue.size() == 0) {
		    synchronized(sendQueueLock) {
		        try {
		            sendQueueLock.wait(10 * 1000);
		        }
		        catch(Exception e) {
			    e.printStackTrace();
		        }
	 	    }
	        }

	        Message m = null;
	        synchronized(queue) {
		    // If we appear to be getting overloaded, send a bunch
		    // at once.
		    if (queue.size() > 5) {
        		// purgeTransientMessages();
			m = new MessageVector((Vector)queue.clone());
			queue.removeAllElements();
		    }
		    else {
		        m = (Message) queue.elementAt(0);
		        queue.removeElementAt(0);
		    }
	        }

	        synchronized(sessionLock) {
                    if (channel != null) {
                        try {
			    if (m.sendToAll())
                               channel.uniformSendToAll(client, m.toData());
			    else
                               channel.uniformSendToOthers(client, m.toData());
                        }
                        catch(Exception sdex) {
                            sdex.printStackTrace();
                        }
	    	     	messageCount++;
		    }
	        }
		if (queue.isEmpty()) {
		    synchronized(shutdownLock) {
		        shutdownLock.notifyAll();
		    }
		    if (waitingForShutdown)
			running = false;
		}
	     }
	}
    }


}


