package sieve.workbench.event;

import java.util.EventObject;
import java.io.Serializable;

/**
 * An EditEvent describes a change made to a WorkBenchComponent. Typically,
 * the PropertyEditEvent subclass will be used to wrap a PropertyChangeEvent.
 * <P>
 * EditEvent differs from PropertyChangeEvent (and similar existing classes)
 * in several ways:
 * <OL>
 * <LI> It includes a method which will apply the change to another
 * object
 * <LI> It includes a method which will determine whether one edit
 * obsoletes another.
 * </OL>
 * <P>
 * The primary goal of this class is to allow Sieve-savvy applications
 * to generate alternatives to PropertyChangeEvent which allow finer-grained
 * changes. In other words, this class allows creation of Sieve-savvy,
 * <I>collaboration-aware</I> components.
 * <P>
 * A future revision of this class could consider adding the ability to
 * undo a change to an object.
 */
public abstract class EditEvent extends EventObject implements Serializable {

    public EditEvent(Object source) {
	super(source);
    }

    /**
     * Applies the change represented by this event to another object.
     * Currently we fail quietly if the target object isn't of the right
     * type or if the change cannot be applied. Eventually we should
     * probably come up with exceptions for these situations, but at the
     * moment it isn't clear how such exceptions would be handled.
     */
    public abstract void applyChange(Object target);

    /**
     * Returns true if the new event obsoletes this event. Consider,
     * for example, a series of EditEvents describing changes to a shared
     * text area. A subsequent EditEvent which included the entire content
     * of the editor would obsolete all of the individual edits.
     */
    public abstract boolean isObsoletedBy(EditEvent newEvent);

    /**
     * This is just like Object.equals() -- it's different because we need to
     * force subclasses to implement it.
     */
    public abstract boolean isEquivalentTo(EditEvent newEvent);
}

